#!/usr/bin/env python3

# This file is part of Shoebot.
# Copyright (C) 2007-2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import sys
from collections import namedtuple
from enum import Enum

from cairo import PATH_CLOSE_PATH, PATH_CURVE_TO, PATH_LINE_TO, PATH_MOVE_TO

from shoebot.core.backend import cairo, driver, gi
from shoebot.util import ShoebotInstallError, _copy_attrs

from .basecolor import ColorMixin
from .bezier import BezierPath
from .grob import Grob

try:
    gi.require_version("Pango", "1.0")
    gi.require_version("PangoCairo", "1.0")
    from gi.repository import Pango, PangoCairo
except ValueError as no_pango:
    global Pango, PangoCairo

    # workaround for readthedocs where Pango is not installed,
    print(_("Pango not found - typography will not be available."), file=sys.stderr)

    class FakePango(object):
        class Weight(Enum):
            # Weights copied from PangoWeight
            THIN = 100
            ULTRALIGHT = 200
            LIGHT = 300
            SEMILIGHT = 350
            BOOK = 380
            NORMAL = 400
            MEDIUM = 500
            SEMIBOLD = 600
            BOLD = 700
            ULTRABOLD = 800
            HEAVY = 900
            ULTRAHEAVY = 1000

        def __getattr__(self, item):
            if item == "Weight":
                return FakePango.Weight

            raise NotImplementedError("FakePango does not implement %s" % item)

    Pango = FakePango()
    PangoCairo = FakePango()


# Pango Utility functions
def pangocairo_create_context(cr):
    """
    Create a PangoCairo context from a given pycairo context.

    If python-gi-cairo is not installed PangoCairo.create_context dies
    with an unhelpful KeyError, output a better error if that happens.
    """
    # TODO move this to core.backend
    try:
        return PangoCairo.create_context(cr)
    except KeyError as e:
        if e.args == ("could not find foreign type Context",):
            raise ShoebotInstallError(
                "Error creating PangoCairo missing dependency: python-gi-cairo"
            )
    raise


def _alignment_name_to_pango(alignment):
    if alignment == "right":
        return Pango.Alignment.RIGHT
    elif alignment == "center":
        return Pango.Alignment.CENTER
    elif alignment == "justify":
        return Pango.Alignment.LEFT

    return Pango.Alignment.LEFT


TextBounds = namedtuple("TextBounds", "x y width height")
TextBounds.__doc__ = """\
Text Bounds in pixels.

:param x: X position in pixels.
:param y: Y position in pixels.
:param width: Width in pixels.
:param height: Height in pixels."""


class Text(Grob, ColorMixin):
    """
    Changes from Nodebox 1:
        font in Nodebox is a native Cocoa font, here it is the font name.
        _fontsize, _fontsize, _lineheight, _align in Nodebox are public fields.

        Implementation of fonts uses Pango instead of Cocoa.
    """

    def __init__(
        self,
        bot,
        text,
        x=0,
        y=0,
        width=None,
        height=None,
        outline=False,
        ctx=None,
        draw=True,
        **kwargs,
    ):
        self._canvas = canvas = bot._canvas
        Grob.__init__(self, bot)
        ColorMixin.__init__(self, **kwargs)

        self.text = str(text)
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.outline = outline

        self.font = kwargs.get("font", canvas.fontfile)
        self.fontsize = kwargs.get("fontsize", canvas.fontsize)
        self.align = kwargs.get("align", canvas.align)
        self.lineheight = kwargs.get("lineheight", canvas.lineheight)
        self.indent = kwargs.get("indent")

        self.hintstyle = kwargs.get("hintstyle", canvas.hintstyle)
        self.hintmetrics = kwargs.get("hintmetrics", canvas.hintmetrics)
        self.antialias = kwargs.get("antialias", canvas.antialias)
        self.subpixelorder = kwargs.get("subpixelorder", canvas.subpixelorder)
        # these are the settings that we have to define with the
        # Pango.Markup hack (see below in _pre_render), so it's neater to
        # have them in a dict
        self.markup_vars = {
            arg: kwargs[arg]
            for arg in (
                "underline",
                "underlinecolor",
                "overline",
                "overlinecolor",
            )
            if arg in kwargs
        }
        if "tracking" in kwargs:
            # different name in Pango
            # and also the value of letter_spacing is in 1/1024ths of
            # a pt, so we multiply that. AND it needs ints
            self.markup_vars["letter_spacing"] = int(1024 * kwargs["tracking"])

        # Setup hidden vars for Cairo / Pango specific bits:
        self._ctx = ctx
        self._pangocairo_ctx = None
        self._pango_fontface = Pango.FontDescription.from_string(self.font)

        # then we set fontsize (multiplied by Pango.SCALE)
        self._pango_fontface.set_absolute_size(self.fontsize * Pango.SCALE)

        # Pre-render some stuff to enable metrics sizing
        self._pre_render()

        if draw:
            # this way we do not render if we only need to create metrics
            if bool(ctx):
                self._render(self._ctx)
            else:
                # Normal rendering, can be deferred
                self._deferred_render()
        self._prerendered = draw

    # pre rendering is needed to measure the metrics of the text, it's also
    # useful to get the path, without the need to call _render()
    def _pre_render(self):
        # we use a new CairoContext to pre render the text
        rs = cairo.RecordingSurface(cairo.CONTENT_ALPHA, None)
        cr = cairo.Context(rs)
        cr = driver.ensure_pycairo_context(cr)

        # apply font options if set
        if self.hintstyle or self.hintmetrics or self.subpixelorder or self.antialias:
            opts = cairo.FontOptions()
            # map values to Cairo constants
            if self.antialias:
                opts.set_antialias(getattr(cairo.Antialias, self.antialias.upper()))
            if self.hintstyle:
                opts.set_hint_style(getattr(cairo.HintStyle, self.hintstyle.upper()))
            if self.hintmetrics:
                opts.set_hint_metrics(
                    getattr(cairo.HintMetrics, self.hintmetrics.upper())
                )
            if self.subpixelorder:
                opts.set_subpixel_order(
                    getattr(cairo.SubpixelOrder, self.subpixelorder.upper())
                )
            cr.set_font_options(opts)

        self._pangocairo_ctx = pangocairo_create_context(cr)
        self._pango_layout = PangoCairo.create_layout(cr)
        # layout line spacing
        # TODO: the behaviour is not the same as nodebox yet
        # self.layout.set_spacing(int(((self.lineheight-1)*self._fontsize)*Pango.SCALE)) #pango requires an int casting
        # we pass pango font description and the text to the pango layout
        self._pango_layout.set_font_description(self._pango_fontface)

        if not self.markup_vars:
            self._pango_layout.set_text(self.text, -1)
            return
        # some of the specified settings require a Pango.Markup hack
        # see https://stackoverflow.com/questions/55533312/how-to-create-a-letter-spacing-attribute-with-pycairo
        # and https://developer.gnome.org/pango/1.46/pango-Markup.html

        # we want to output something like
        # <span letter_spacing="2048">Hello World</span>
        markup_styles = " ".join(
            [f'{setting}="{value}"' for setting, value in self.markup_vars.items()]
        )
        self._pango_layout.set_markup(f"<span {markup_styles}>{self.text}</span>")

        # check if max text width is set and pass it to pango layout
        # text will wrap, meanwhile it checks if and indent has to be applied
        # indent is subordinated to width because it makes no sense on a single-line text block
        if self.width:
            self._pango_layout.set_width(int(self.width) * Pango.SCALE)
            if self.indent:
                self._pango_layout.set_indent(self.indent * Pango.SCALE)
        # set text alignment
        self._pango_layout.set_alignment(_alignment_name_to_pango(self.align))
        if self.align == "justify":
            self._pango_layout.set_justify(True)

    def _get_context(self):
        self._ctx = self._ctx or cairo.Context(
            cairo.RecordingSurface(cairo.CONTENT_ALPHA, None)
        )
        return self._ctx

    def _render(self, ctx=None):
        if not self._prerendered:
            return
        ctx = ctx or self._get_context()
        pycairo_ctx = driver.ensure_pycairo_context(ctx)
        # we build a PangoCairo context linked to cairo context
        # then we create a pango layout.
        # Update the context as we already used a null one on the pre-rendering
        # supposedly there should not be a big performance penalty
        self._pangocairo_ctx = pangocairo_create_context(pycairo_ctx)

        if self._fillcolor is None:
            return

        # Go to initial point (CORNER or CENTER):
        transform = self._call_transform_mode(self._transform)
        ctx.set_matrix(transform)
        ctx.translate(self.x, self.y - self.baseline)

        if not self.outline:
            # In outline, the caller will stroke the generated path
            ctx.set_source_rgba(*self._fillcolor)
        PangoCairo.show_layout(pycairo_ctx, self._pango_layout)
        PangoCairo.update_layout(pycairo_ctx, self._pango_layout)

    # This version is probably more pangoesque, but the layout iterator
    # caused segfaults on some system
    @property
    def baseline(self):
        self.iter = self._pango_layout.get_iter()
        baseline_y = self.iter.get_baseline()
        baseline_delta = baseline_y / Pango.SCALE
        return baseline_delta

    #    @property
    #    def baseline(self):
    #        # retrieves first line of text block
    #        first_line = self.layout.get_line(0)
    #        # get the logical extents rectangle of first line
    #        first_line_extent = first_line.get_extents()[1]
    #        # get the descent value, in order to calculate baseline position
    #        first_line_descent = Pango.DESCENT(first_line.get_extents()[1])
    #        # gets the baseline offset from the top of thext block
    #        baseline_delta = (first_line_extent[3]-first_line_descent)/Pango.SCALE
    #        return (baseline_delta)

    @property
    def metrics(self):
        w, h = self._pango_layout.get_pixel_size()
        return w, h

    @property
    def bounds(self):
        """
        :return: TextBounds namedtuple containing bounds as (x, y, width, height)
        """
        bounds_rect, _ = self._pango_layout.get_pixel_extents()
        return TextBounds(
            bounds_rect.x + self.x,
            bounds_rect.y + self.y - self.baseline,
            bounds_rect.width,
            bounds_rect.height,
        )

    # this function is quite computational expensive
    # there should be a way to make it faster, by not creating a new context each time it's called

    @property
    def path(self):
        if not self._pangocairo_ctx:
            self._pre_render()

        # Render path data to a temporary cairo Context
        cairo_ctx = cairo.Context(cairo.RecordingSurface(cairo.CONTENT_ALPHA, None))
        cairo_ctx = driver.ensure_pycairo_context(cairo_ctx)
        cairo_ctx.move_to(self.x, self.y - self.baseline)
        # show_layout should work here, but fills the path instead,
        # instead, use layout_path to render the layout.
        PangoCairo.layout_path(cairo_ctx, self._pango_layout)

        # Parse cairo path into Shoebot BezierPath
        cairo_text_path = cairo_ctx.copy_path()
        p = BezierPath(self._bot)
        for item in cairo_text_path:
            cmd = item[0]
            args = item[1]
            if cmd == PATH_MOVE_TO:
                p.moveto(*args)
            elif cmd == PATH_LINE_TO:
                p.lineto(*args)
            elif cmd == PATH_CURVE_TO:
                p.curveto(*args)
            elif cmd == PATH_CLOSE_PATH:
                p.closepath()

        del cairo_ctx
        return p

    def _get_center(self):
        """Returns the center point of the path, disregarding transforms."""
        w, h = self._pango_layout.get_pixel_size()
        x = self.x + w / 2
        y = self.y + h / 2
        return x, y

    center = property(_get_center)

    def copy(self):
        copied_instance = self.__class__(self._bot, self.text)
        _copy_attrs(
            self,
            copied_instance,
            (
                "x",
                "y",
                "width",
                "height",
                "_fillcolor",
                "fontfile",
                "fontsize",
                "align",
                "lineheight",
                "_transform",
                "_transformmode",
            ),
        )
        return copied_instance



#!/usr/bin/env python3

from . import nodebox

__all__ = [
    "nodebox",
    "bot",
    "input_device",
    "InputDeviceMixin",
    "VarListener",
]

from .bot import Bot
from .nodebox import NodeBot
from shoebot.core.var_listener import VarListener



"""
Cairo implementation and GObject preference can be set by channging the environment variable SHOEBOT_GRAPHICS.

Cairo:

Shoebot Core [cairo|cairocffi]
Shoebot GUI  [cairo (can convert from cairocffi]

GObject:

GI  - Works for everything
PGI - Text does not work (because of Pango issues).


Environment variable:

Set SHOEBOT_GRAPHICS environment variable to change the order Shoebot will try import graphics modules.

SHOEBOT_GRAPHICS='cairo=cairocffi,cairo gi=pgi'

There is currently no output if Shoebot can't use the chosen implementation.


Check settings

Use setup.py diagnose to output the current graphics settings.
"""

import os
import sys


class BackendMixin(object):
    """
    Mixin to abstract different implementations of the same library.
    """

    def import_libs(self, module_names, impl_name):
        """
        Loop through module_names,
        add has_.... booleans to class
        set ..._impl to first successful import

        :param module_names:  list of module names to try importing
        :param impl_name:  used in error output if no modules succeed
        :return: name, module from first successful implementation
        """
        for name in module_names:
            try:
                module = __import__(name)
                has_module = True
            except ImportError:
                module = None
                has_module = False
            setattr(self, name, module)
            setattr(self, "has_%s" % name, has_module)

        for name in module_names:
            try:
                return name, __import__(name)
            except ImportError:
                pass
        raise ImportError(
            "No %s Implementation found, tried: %s"
            % (impl_name, " ".join(module_names))
        )

    def get_libs(self):
        return {}


def sort_by_preference(options, prefer):
    """
    :param options: List of options
    :param prefer: Prefered options
    :return:

    Pass in a list of options, return options in 'prefer' first

    >>> sort_by_preference(["cairo", "cairocffi"], ["cairocffi"])
    ["cairocffi", "cairo"]
    """
    if not prefer:
        return options
    return sorted(options, key=lambda x: (prefer + options).index(x))


class CairoGIBackend(BackendMixin):
    """
    Graphics backend using gi.repository or pgi
    PyCairo / CairoCFFI (+PyCairo needed if using Gtk Too)
    """

    def __init__(self, options):
        cairo_pref = sort_by_preference(
            ["cairo", "cairocffi"], options.get("cairo", "").split(",")
        )
        gi_pref = sort_by_preference(["gi", "pgi"], options.get("gi", "").split(","))
        self.cairo_lib, self.cairo_module = self.import_libs(cairo_pref, "Cairo")
        self.gi_lib, self.gi_module = self.setup_gi(gi_pref)

    def get_libs(self):
        return {"gi": self.gi_lib, "cairo": self.cairo_lib}

    def setup_gi(self, gi_pref):
        name, gi_module = self.import_libs(["gi", "pgi"], "gi")
        if name == "pgi":
            gi_module.install_as_gi()
        return name, gi_module

    def ensure_pycairo_context(self, ctx):
        """
        If ctx is a cairocffi Context convert it to a PyCairo Context
        otherwise return the original context

        :param ctx:
        :return:
        """
        if self.cairocffi and isinstance(ctx, self.cairocffi.Context):
            from shoebot.util.cairocffi.cairocffi_to_pycairo import (
                _UNSAFE_cairocffi_context_to_pycairo,
            )

            return _UNSAFE_cairocffi_context_to_pycairo(ctx)
        else:
            return ctx


def get_driver_options():
    """
    Interpret env var as key=value
    :return:
    """
    options = os.environ.get("SHOEBOT_GRAPHICS")
    if not options:
        return {}

    try:
        return dict([kv.split("=") for kv in options.split()])
    except ValueError:
        sys.stderr.write("Bad option format.\n")
        sys.stderr.write(
            "Environment variable should be in the format key=value separated by spaces.\n\n"
        )
        sys.stderr.write("SHOEBOT_GRAPHICS='cairo=cairocffi,cairo gi=pgi'\n")
        sys.exit(1)


driver = CairoGIBackend(get_driver_options())
gi = driver.gi_module
cairo = driver.cairo_module



# Copyright (c) 2007 Tom De Smedt.
# See LICENSE.txt for details.

from math import degrees, sqrt, atan2
from math import radians, sin, cos

CORNER = "corner"
CENTER = "center"

DEFAULT = "default"
HIGHLIGHT = "highlight"  # used to mark shortest paths in pink
ROOT = "root"  # used for the root node, big pink text
LIGHT = "light"  # slightly important nodes (these are a bit darker)
DARK = "dark"  # very important nodes (these are blue)
BACK = "back"  # used as "back-button", green with a curved edge
IMPORTANT = "important"  # like dark, but with a double stroke
MARKED = "marked"  # has a white dot inside the node

#### GRAPH STYLES ####################################################################################


class styles(dict):
    def __init__(self, graph):
        self.guide = styleguide(graph)

    def apply(self):
        self.guide.apply()

    def create(self, stylename, **kwargs):
        """
        Creates a new style which inherits from the default style,
        or any other style which name is supplied to the optional template parameter.
        """
        if stylename == "default":
            self[stylename] = style(stylename, self._ctx, **kwargs)
            return self[stylename]
        k = kwargs.get("template", "default")
        s = self[stylename] = self[k].copy(stylename)
        for attr in kwargs:
            if attr in s.__dict__:
                s.__dict__[attr] = kwargs[attr]
        return s

    def append(self, style):
        self[style.name] = style

    def __getattr__(self, a):
        """
        Keys in the dictionaries are accessible as attributes.
        """
        if a in self:
            return self[a]
        raise AttributeError("'styles' object has no attribute '" + a + "'")

    def __setattr__(self, a, v):
        """
        Setting an attribute is like setting it in all of the contained styles.
        """
        if a == "guide":
            self.__dict__["guide"] = v
        elif len(self) > 0 and a in list(self.values())[0].__dict__:
            for style in list(self.values()):
                style.__dict__[a] = v
        else:
            raise AttributeError("'style' object has no attribute '" + a + "'")

    def copy(self, graph):
        """
        Returns a copy of all styles and a copy of the styleguide.
        """
        s = styles(graph)
        s.guide = self.guide.copy(graph)
        dict.__init__(s, [(v.name, v.copy()) for v in list(self.values())])
        return s


#### GRAPH STYLE GUIDE ###############################################################################
# Each node gets the default colors, type and drawing functions.
# The guide defines how and when to apply other styles based on node properties.
# It contains a set of style name keys linked to x(graph, node) functions.
# If such a function returns True for a node, the style is applied to that node.


class styleguide(dict):
    def __init__(self, graph):
        self.graph = graph
        self.order = []

    def append(self, stylename, function):
        """
        The name of a style and a function that takes a graph and a node.
        It returns True when the style should be applied to the given node.
        """
        self[stylename] = function

    def clear(self):
        self.order = []
        dict.__init__(self)

    def apply(self):
        """
        Check the rules for each node in the graph and apply the style.
        """
        sorted = self.order + list(self.keys())
        unique = []
        [unique.append(x) for x in sorted if x not in unique]
        for node in self.graph.nodes:
            for s in unique:
                if s in self and self[s](self.graph, node):
                    node.style = s

    def copy(self, graph):
        """
        Returns a copy of the styleguide for the given graph.
        """
        g = styleguide(graph)
        g.order = self.order
        dict.__init__(g, [(k, v) for k, v in self.items()])
        return g


#### GRAPH STYLE #####################################################################################


class style:
    def __init__(self, name, _ctx, **kwargs):
        """
        Graph styling. 
        The default style is used for edges.
        When text is set to None, no id label is displayed.
        """
        self.name = name
        self._ctx = _ctx
        if not _ctx:
            return

        # Defaults for colors and typography.
        # self.background  = _ctx.color(0.18, 0.23, 0.28, 1.00)
        self.background = _ctx.color(0.45, 0.45, 0.5, 1.00)
        self.traffic = _ctx.color(0.00, 0.00, 0.00, 0.07)
        self.fill = _ctx.color(0.00, 0.00, 0.00, 0.10)
        self.stroke = _ctx.color(0.80, 0.80, 0.80, 0.75)
        self.strokewidth = 0.5
        self.text = _ctx.color(1.00, 1.00, 1.00, 0.85)
        self.font = "Calibri"
        self.fontsize = 10
        self.textwidth = 100
        self.align = 1
        self.depth = True

        # The actual drawing methods are just a bunch of monkey patches,
        # so another function can easily be assigned.
        # Call style.draw_method(style, params) instead of style.draw_method(params).
        self.graph_background = graph_background
        self.graph_traffic = graph_traffic
        self.node = node
        self.node_label = node_label
        self.edges = edges
        self.edge = edge
        self.edge_arrow = edge_arrow
        self.edge_label = edge_label
        self.path = path

        # Each of the attributes is an optional named parameter in __init__().
        for attr in kwargs:
            if attr in self.__dict__:
                self.__dict__[attr] = kwargs[attr]

        # Use the Colors library for gradients and shadows?
        if self.depth:
            try:
                global colors
                colors = _ctx.ximport("colors")
            except:
                self.depth = False

    def copy(self, name=None):
        # Copy all attributes, link all monkey patch methods.
        s = style(self.name, self._ctx)
        for attr in self.__dict__:
            v = self.__dict__[attr]
            if isinstance(v, self.fill.__class__):
                v = v.copy()
            s.__dict__[attr] = v
        if name is not None:
            s.name = name

        return s


# --- GRAPH BACKGROUND --------------------------------------------------------------------------------


def graph_background(s):
    """
    Graph background color.
    """
    s._ctx.background(s.background)

    if s.depth:
        try:
            clr = colors.color(s.background).darker(0.2)
            p = s._ctx.rect(0, 0, s._ctx.WIDTH, s._ctx.HEIGHT, draw=False)
            colors.gradientfill(p, clr, clr.lighter(0.35))
            colors.shadow(dx=0, dy=0, blur=2, alpha=0.935, clr=s.background)
        except:
            pass


# --- GRAPH TRAFFIC -----------------------------------------------------------------------------------


def graph_traffic(s, node, alpha=1.0):
    """
    Visualization of traffic-intensive nodes (based on their centrality).
    """
    r = node.__class__(None).r
    r += (node.weight + 0.5) * r * 5
    s._ctx.nostroke()
    if s.traffic:
        s._ctx.fill(s.traffic.r, s.traffic.g, s.traffic.b, s.traffic.a * alpha)
        s._ctx.oval(node.x - r, node.y - r, r * 2, r * 2)


# --- NODE --------------------------------------------------------------------------------------------


def node(s, node, alpha=1.0):
    """
    Visualization of a default node.
    """
    if s.depth:
        try:
            colors.shadow(dx=5, dy=5, blur=10, alpha=0.5 * alpha)
        except:
            pass

    s._ctx.nofill()
    s._ctx.nostroke()
    if s.fill:
        s._ctx.fill(s.fill.r, s.fill.g, s.fill.b, s.fill.a * alpha)
    if s.stroke:
        s._ctx.strokewidth(s.strokewidth)
        s._ctx.stroke(s.stroke.r, s.stroke.g, s.stroke.b, s.stroke.a * alpha * 3)
    r = node.r
    s._ctx.oval(node.x - r, node.y - r, r * 2, r * 2)


# --- NODE LABEL -------------------------------------------------------------------------------------


def node_label(s, node, alpha=1.0):
    """
    Visualization of a node's id.
    """
    if s.text:
        # s._ctx.lineheight(1)
        s._ctx.font(s.font)
        s._ctx.fontsize(s.fontsize)
        s._ctx.nostroke()
        s._ctx.fill(s.text.r, s.text.g, s.text.b, s.text.a * alpha)

        # Cache an outlined label text and translate it.
        # This enhances the speed and avoids wiggling text.
        try:
            p = node._textpath
        except:
            txt = node.label
            try:
                txt = str(txt)
            except:
                try:
                    txt = txt.decode("utf-8")
                except:
                    pass
            # Abbreviation.
            # root = node.graph.root
            # if txt != root and txt[-len(root):] == root:
            #    txt = txt[:len(txt)-len(root)]+root[0]+"."
            dx, dy = 0, 0
            if s.align == 2:  # CENTER
                dx = -s._ctx.textwidth(txt, s.textwidth) / 2
                dy = s._ctx.textheight(txt) / 2
            node._textpath = s._ctx.textpath(txt, dx, dy, width=s.textwidth)
            p = node._textpath

        if s.depth:
            try:
                __colors.shadow(dx=2, dy=4, blur=5, alpha=0.3 * alpha)
            except:
                pass

        s._ctx.push()
        s._ctx.translate(node.x, node.y)
        s._ctx.scale(alpha)
        s._ctx.drawpath(p.copy())
        s._ctx.pop()


# --- EDGES -------------------------------------------------------------------------------------------


def edges(s, edges, alpha=1.0, weighted=False, directed=False):
    """
    Visualization of the edges in a network.
    """
    p = s._ctx.BezierPath()

    if directed and s.stroke:
        pd = s._ctx.BezierPath()
    if weighted and s.fill:
        pw = [s._ctx.BezierPath() for i in range(11)]

    # Draw the edges in a single BezierPath for speed.
    # Weighted edges are divided into ten BezierPaths,
    # depending on their weight rounded between 0 and 10.
    if len(edges) == 0:
        return
    for e in edges:
        try:
            s2 = e.node1.graph.styles[e.node1.style]
        except:
            s2 = s
        if s2.edge:
            s2.edge(s2, p, e, alpha)
            if directed and s.stroke:
                s2.edge_arrow(s2, pd, e, radius=10)
            if weighted and s.fill:
                s2.edge(s2, pw[int(e.weight * 10)], e, alpha)

    s._ctx.autoclosepath(False)
    s._ctx.nofill()
    s._ctx.nostroke()

    # All weighted edges use the default fill.
    if weighted and s.fill:
        r = e.node1.__class__(None).r
        s._ctx.stroke(s.fill.r, s.fill.g, s.fill.b, s.fill.a * 0.65 * alpha)
        for w in range(1, len(pw)):
            s._ctx.strokewidth(r * w * 0.1)
            s._ctx.drawpath(pw[w].copy())

    # All edges use the default stroke.
    if s.stroke:
        s._ctx.strokewidth(s.strokewidth)

        s._ctx.stroke(s.stroke.r, s.stroke.g, s.stroke.b, s.stroke.a * 0.65 * alpha)

    s._ctx.drawpath(p.copy())

    if directed and s.stroke:
        # clr = s._ctx.stroke().copy()
        clr = s._ctx.color(
            s.stroke.r, s.stroke.g, s.stroke.b, s.stroke.a * 0.65 * alpha
        )

        clr.a *= 1.3

        s._ctx.stroke(clr)
        s._ctx.drawpath(pd.copy())

    for e in edges:
        try:
            s2 = self.styles[e.node1.style]
        except:
            s2 = s
        if s2.edge_label:
            s2.edge_label(s2, e, alpha)


# --- EDGE --------------------------------------------------------------------------------------------


def edge(s, path, edge, alpha=1.0):
    """
    Visualization of a single edge between two nodes.
    """
    path.moveto(edge.node1.x, edge.node1.y)
    if edge.node2.style == BACK:
        path.curveto(
            edge.node1.x,
            edge.node2.y,
            edge.node2.x,
            edge.node2.y,
            edge.node2.x,
            edge.node2.y,
        )
    else:
        path.lineto(edge.node2.x, edge.node2.y)


# --- EDGE ARROW --------------------------------------------------------------------------------------


def edge_arrow(s, path, edge, radius):

    if edge.node2.style == BACK:
        return

    x0, y0 = edge.node1.x, edge.node1.y
    x1, y1 = edge.node2.x, edge.node2.y

    coordinates = lambda x, y, d, a: (x + cos(radians(a)) * d, y + sin(radians(a)) * d)

    # Find the edge's angle based on node1 and node2 position.
    a = degrees(atan2(y1 - y0, x1 - x0))

    # The arrow points to node2's rim instead of it's center.
    r = edge.node2.r
    d = sqrt(pow(x1 - x0, 2) + pow(y1 - y0, 2))
    x01, y01 = coordinates(x0, y0, d - r - 1, a)

    # Find the two other arrow corners under the given angle.
    r = edge.node1.r
    r = radius
    dx1, dy1 = coordinates(x01, y01, -r, a - 20)
    dx2, dy2 = coordinates(x01, y01, -r, a + 20)

    path.moveto(x01, y01)
    path.lineto(dx1, dy1)
    path.lineto(dx2, dy2)
    path.lineto(x01, y01)
    path.moveto(x1, y1)


# --- EDGE LABEL --------------------------------------------------------------------------------------


def edge_label(s, edge, alpha=1.0):
    """
    Visualization of the label accompanying an edge.
    """
    if s.text and edge.label != "":
        s._ctx.nostroke()
        s._ctx.fill(s.text.r, s.text.g, s.text.b, s.text.a * alpha * 0.75)
        s._ctx.lineheight(1)
        s._ctx.font(s.font)
        s._ctx.fontsize(s.fontsize * 0.75)

        # Cache an outlined label text and translate it.
        # This enhances the speed and avoids wiggling text.
        try:
            p = edge._textpath
        except:
            try:
                txt = str(edge.label)
            except:
                try:
                    txt = edge.label.decode("utf-8")
                except:
                    pass
            edge._textpath = s._ctx.textpath(
                txt, s._ctx.textwidth(" "), 0, width=s.textwidth
            )
            p = edge._textpath

        # Position the label centrally along the edge line.
        a = degrees(atan2(edge.node2.y - edge.node1.y, edge.node2.x - edge.node1.x))
        d = sqrt(
            (edge.node2.x - edge.node1.x) ** 2 + (edge.node2.y - edge.node1.y) ** 2
        )
        d = abs(d - s._ctx.textwidth(edge.label)) * 0.5

        s._ctx.push()
        s._ctx.transform(CORNER)
        s._ctx.translate(edge.node1.x, edge.node1.y)
        s._ctx.rotate(-a)
        s._ctx.translate(d, s.fontsize * 1.0)
        s._ctx.scale(alpha)

        # Flip labels on the left hand side so they are legible.
        if 90 < a % 360 < 270:
            s._ctx.translate(s._ctx.textwidth(edge.label), -s.fontsize * 2.0)
            s._ctx.transform(CENTER)
            s._ctx.rotate(180)
            s._ctx.transform(CORNER)

        s._ctx.drawpath(p.copy())
        s._ctx.pop()


# ---- PATH -------------------------------------------------------------------------------------------


def path(s, graph, path):
    """
    Visualization of a shortest path between two nodes.
    """

    def end(n):
        r = n.r * 0.35
        s._ctx.oval(n.x - r, n.y - r, r * 2, r * 2)

    if path and len(path) > 1 and s.stroke:

        s._ctx.nofill()
        s._ctx.stroke(s.stroke.r, s.stroke.g, s.stroke.b, s.stroke.a)
        if s.name != DEFAULT:
            s._ctx.strokewidth(s.strokewidth)
        else:
            s._ctx.strokewidth(s.strokewidth * 2)

        first = True
        for id in path:
            n = graph[id]
            if first:
                first = False
                s._ctx.beginpath(n.x, n.y)
                end(n)
            else:
                s._ctx.lineto(n.x, n.y)
        s._ctx.endpath()
        end(n)



#!/usr/bin/env python3

# This file is part of Shoebot.
# Copyright (C) 2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Assorted utility functions, mainly for color and font handling"""


class ShoebotInstallError(Exception):
    pass


class UnbufferedFile:
    """
    File wrapper, that flushes on writes.

    http://stackoverflow.com/questions/230751/how-to-flush-output-of-python-print
    """

    def __init__(self, fd):
        self.fd = fd

    def write(self, x):
        self.fd.write(x)
        self.fd.flush()

    def writelines(self, lines):
        self.fd.writelines(lines)
        self.fd.flush()

    def flush(self):
        return self.fd.flush()

    def close(self):
        return self.fd.close()

    def fileno(self):
        return self.fd.fileno()


def _copy_attr(v):
    if v is None:
        return None
    elif hasattr(v, "copy"):
        return v.copy()
    elif isinstance(v, list):
        return list(v)
    elif isinstance(v, tuple):
        return tuple(v)
    elif isinstance(v, (int, str, float, bool)):
        return v
    else:
        raise ValueError(_("Don't know how to copy '%s'.") % v)


def _copy_attrs(source, target, attrs):
    """
    Copy attributes from source to target.

    :param source: source object
    :param target: destination object
    :param attrs: sequence of attributes to copy.
    :return:
    """
    for attr in attrs:
        setattr(target, attr, _copy_attr(getattr(source, attr)))



"""
The pubsub library enables communication between the GUI parts of shoebot
and the commandline shell.
"""

import collections
import pubsub

# Only the 'shoebot' channel is used
channel = pubsub.subscribe("shoebot")


def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = collections.namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None,) * len(T._fields)
    if isinstance(default_values, collections.abc.Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    return T


Event = namedtuple = namedtuple_with_defaults("Event", "type data", dict(data=None))

QUIT_EVENT = "quit"
SOURCE_CHANGED_EVENT = "source-changed"
VARIABLE_CHANGED_EVENT = "variable-updated"
SET_WINDOW_TITLE_EVENT = "set-window-title"
REDRAW_EVENT = "redraw"


def next_event(block=False, timeout=None):
    """
    Get the next available event or None

    :param block:
    :param timeout:
    :return: None or (event, data)
    """
    try:
        return next(channel.listen(block=block, timeout=timeout))["data"]
    except StopIteration:
        return None


def event_is(event, event_t):
    """
    Check if event type
    :param event:   event to compare
    :param event_t: event type
    :return: bool
    """
    return event is not None and event.type == event_t


def publish_event(event_t, data=None, extra_channels=None, wait=None):
    """
    Publish an event ot any subscribers.

    :param event_t:  event type
    :param data:     event data
    :param extra_channels:
    :param wait:
    :return:
    """
    event = Event(event_t, data)
    pubsub.publish("shoebot", event)
    for channel_name in extra_channels or []:
        pubsub.publish(channel_name, event)
    if wait is not None:
        channel = pubsub.subscribe(wait)
        channel.listen(wait)
        channel.unsubscribe(wait)



#!/usr/bin/env python

# This file is part of Shoebot.
# Copyright (C) 2007-2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import signal
import sys
import threading

from shoebot.core.backend import cairo
from shoebot.grammar import NodeBot

# TODO - Check if this needs importing here:
# from shoebot.data import MOVETO, RMOVETO, LINETO, RLINETO, CURVETO, RCURVETO, ARC, ELLIPSE, CLOSE, LEFT, RIGHT, ShoebotError, ShoebotScriptError
from time import sleep

from shoebot.core.events import publish_event, QUIT_EVENT

RGB = "rgb"
HSB = "hsb"

CENTER = "center"
CORNER = "corner"
CORNERS = "corners"

NODEBOX = "nodebox"


def _save():
    # Dummy function used by color lib; TODO investigate if this can be implemented
    pass


def _restore():
    # Dummy function used by color lib; TODO investigate if this can be implemented
    pass


# Convenience functions to create a bot, its canvas and sink


def create_canvas(
    src,
    format=None,
    outputfile=None,
    multifile=False,
    buff=None,
    window=False,
    title=None,
    fullscreen=None,
    show_vars=False,
):
    """
    Create canvas and sink for attachment to a bot

    canvas is what draws images, 'sink' is the final consumer of the images

    :param src: Defaults for title or outputfile if not specified.

    :param format: CairoImageSink image format, if using buff instead of outputfile
    :param buff: CairoImageSink buffer object to send output to

    :param outputfile: CairoImageSink output filename e.g. "hello.svg"
    :param multifile: CairoImageSink if True,

    :param title: ShoebotWindow - set window title
    :param fullscreen: ShoebotWindow - set window title
    :param show_vars: ShoebotWindow - display variable window

    Two kinds of sink are provided: CairoImageSink and ShoebotWindow

    ShoebotWindow

    Displays a window to draw shoebot inside.


    CairoImageSink

    Output to a filename (or files if multifile is set), or a buffer object.
    """
    from shoebot.core import (
        CairoCanvas,
        CairoImageSink,
    )  # https://github.com/shoebot/shoebot/issues/206

    if window or show_vars:
        from shoebot.gui import ShoebotWindow

        if not title:
            if src and os.path.isfile(src):
                title = os.path.splitext(os.path.basename(src))[0] + " - Shoebot"
            else:
                title = "Untitled - Shoebot"
        sink = ShoebotWindow(
            title, show_vars, fullscreen=fullscreen, outputfile=outputfile
        )
    elif outputfile:
        sink = CairoImageSink(outputfile, format, multifile, buff)
    else:
        if src and isinstance(src, cairo.Surface):
            outputfile = src
            format = "surface"
        elif src and os.path.isfile(src):
            outputfile = (
                os.path.splitext(os.path.basename(src))[0] + "." + (format or "svg")
            )
        else:
            outputfile = "output.svg"
        sink = CairoImageSink(outputfile, format, multifile, buff)
    canvas = CairoCanvas(sink)

    return canvas


def create_bot(
    src=None,
    format=None,
    outputfile=None,
    iterations=1,
    buff=None,
    window=False,
    title=None,
    fullscreen=None,
    server=False,
    port=7777,
    show_vars=False,
    vars=None,
    namespace=None,
):
    """
    Create a canvas and a bot with the same canvas attached to it

    bot parameters
    :param vars: preset dictionary of vars from the called

    canvas parameters:
    ... everything else ...

    See create_canvas for details on those parameters.

    """
    multifile = True if iterations and iterations > 1 else False
    canvas = create_canvas(
        src=src,
        format=format,
        outputfile=outputfile,
        multifile=multifile,
        buff=buff,
        window=window,
        title=title,
        fullscreen=fullscreen,
        show_vars=show_vars,
    )

    bot = NodeBot(canvas, namespace=namespace, vars=vars)

    if server:
        from shoebot.sbio import SocketServer

        socket_server = SocketServer(bot, "", port=port)
    return bot


class ShoebotThread(threading.Thread):
    """
    Run shoebot in an alternate thread.

    This way the commandline shell can run on the main thread
    and the GUI in a seperate thread without readline
    blocking it.
    """

    def __init__(
        self, create_args, create_kwargs, run_args, run_kwargs, send_sigint=False
    ):
        """
        :param create_args: passed to create_bot
        :param create_kwargs: passed to create_bot
        :param run_args: passed to bot.run
        :param run_kwargs: passed to bot.run
        :param send_sigint: if True then SIGINT will be sent on bot completion
                            so the main thread can terminate
        """
        super(ShoebotThread, self).__init__(daemon=True)
        # isSet() will return True once the bot has been created
        self.bot_ready = threading.Event()

        self.create_args = create_args
        self.create_kwargs = create_kwargs

        self.run_args = run_args
        self.run_kwargs = run_kwargs

        self.success = None  # TODO: fix naming, should probably be thread Event so it can be waited on (and have another name)

        self.send_sigint = send_sigint
        self._sbot = None

    def run(self):
        try:
            sbot = create_bot(*self.create_args, **self.create_kwargs)

            self._sbot = sbot
            self.bot_ready.set()
            success = sbot.run(*self.run_args, **self.run_kwargs)
            self.success = success
        except Exception:
            self.success = False
            print("Exception in shoebot code")
            raise
        finally:
            self.bot_ready.set()  # Stop waiting
            if self.send_sigint:
                os.kill(os.getpid(), signal.SIGINT)

    @property
    def sbot(self):
        """
        :return: bot instance for communication
        """
        self.bot_ready.wait()
        return self._sbot


def run(
    src,
    format=None,
    outputfile=None,
    max_iterations=1,
    buff=None,
    window=True,
    title=None,
    fullscreen=None,
    close_window=False,
    server=False,
    port=7777,
    show_vars=False,
    vars=None,
    namespace=None,
    run_shell=False,
    args=[],
    verbose=False,
    background_thread=True,
):
    """
    Create and run a bot, the arguments all correspond to sanitized
    commandline options.

    :param background_thread: If True then use a background thread.


    Other args are split into create_args and run_args

    See create_bot for details on create_args

    run_args are passed to bot.run - see Nodebot.run



    Background thread:

    readline in python is blocking, running the app in a background
    thread opens up the main thread for IO on stdin/stdout, which
    can be used for communication with shoebot when livecoding is
    enabled.

    See shoebot.io for implementation of the shell, and the gedit
    plugin for an example of using livecoding.
    """
    # Munge shoebogt sys.argv
    sys.argv = [
        sys.argv[0]
    ] + args  # Remove shoebot parameters so sbot can be used in place of the python interpreter (e.g. for sphinx).

    # arguments for create_bot
    create_args = [
        src,
        format,
        outputfile,
        max_iterations,
        buff,
        window,
        title,
        fullscreen,
        server,
        port,
        show_vars,
    ]
    create_kwargs = dict(vars=vars, namespace=namespace)
    run_args = [src]
    run_kwargs = dict(
        max_iterations=max_iterations,
        frame_limiter=window,
        verbose=verbose,
        # run forever except 1. windowed mode is off 2. if --close-window was specified and
        # 3. if an output file was indicated
        run_forever=window and not (close_window or bool(outputfile)),
    )

    # Run shoebot in a background thread so we can run a cmdline shell in the current thread
    if background_thread:
        sbot_thread = ShoebotThread(
            create_args=create_args,
            create_kwargs=create_kwargs,
            run_args=run_args,
            run_kwargs=run_kwargs,
            send_sigint=run_shell,
        )
        sbot_thread.start()
        sbot = sbot_thread.sbot
    else:
        # This is a debug option, things should always work using the
        # background thread (crosses fingers)
        if run_shell:
            # python readline is blocking, so ui must run in a seperate
            # thread
            raise ValueError(
                "UI Must run in a separate thread to shell and shell needs main thread"
            )

        sbot_thread = None
        sbot = create_bot(*create_args, **create_kwargs)
        return sbot.run(*run_args, **run_kwargs)

    if run_shell:
        import shoebot.sbio.shell

        shell = shoebot.sbio.shell.ShoebotCmd(sbot, trusted=True)
        try:
            shell.cmdloop()
        except KeyboardInterrupt as e:
            publish_event(QUIT_EVENT)  # Handle Ctrl-C
            # KeyboardInterrupt is generated by os.kill from the other thread
            if verbose:
                raise
            return False
    elif background_thread:
        try:
            while sbot_thread.is_alive():
                sleep(0.001)
        except KeyboardInterrupt:
            sbot_thread.send_sigint = False
            publish_event(QUIT_EVENT)
            raise
        return sbot_thread.success



"""
Display information to help diagnose install issues.

Currently shows
- OS info
- Python info
- Gtk3 availability (gi.repository vs pgi)

This can be complemented by running the unittests.
"""


import os
import platform
import sys
import traceback

from collections import namedtuple

COL_WIDTH = 10

AvailableModules = namedtuple("AvailableModules", "gi pgi meta pango pubsub rsvg vext")


def display_platform():
    # environment info
    is_virtualenv = "VIRTUAL_ENV" in os.environ

    # operating system info
    def linux_distribution():
        try:
            return platform.linux_distribution()
        except:
            return "N/A"

    # operating system info
    def platform_dist():
        try:
            return platform.dist()
        except:
            return "N/A"

    print(
        """Python:
    sys.executable: %s
    virtualenv: %s
    version: %s
    dist: %s
    linux_distribution: %s
    system: %s
    machine: %s
    platform: %s
    version: %s
    mac_ver: %s
    win32_ver: %s
    """
        % (
            sys.executable,
            is_virtualenv or "no",
            " ".join(sys.version.split("\n")),
            str(" ".join(platform_dist())),
            " ".join(linux_distribution()),
            platform.system(),
            platform.machine(),
            platform.platform(),
            platform.version(),
            platform.mac_ver(),
            platform.win32_ver(),
        )
    )


def import_success_message(module, name):
    return "\n".join(
        [
            "    import %s [success]:" % name.ljust(COL_WIDTH),
            "        " + module.__file__,
        ]
    )


def import_fail_message(mn, reason):
    return "\n".join(
        ["    import %s [failed]:" % mn.ljust(COL_WIDTH), "        " + reason]
    )


def test_import(name, failmsg=None, gi_require=None, gi=None):
    if all([gi_require, gi]):
        try:
            gi.require_version(*gi_require)
        except ValueError as ex:
            # import will be attempted anyway
            err = ex.args[0]
            print(import_fail_message(name, err))
            return
    try:
        module = __import__(name)
        print(import_success_message(module, name))
        return module
    except ImportError as e:
        print(import_fail_message(name, failmsg or str(e)))
    except Exception as e:
        print(import_fail_message(name, str(e)))


def test_imports():
    """
    Attempt to import dependencies.
    """
    print("Test Imports:")
    # gtk
    gi = test_import("gi")
    pgi = test_import("pgi")
    _gi = gi or pgi
    if gi:
        pango = test_import("gi.repository.Pango", gi_require=("Pango", "1.0"), gi=_gi)
    else:
        print("    No gi.repository.Pango implementation, text will not be available")
        pango = None
    # virtualenv help
    vext = test_import("vext")

    # internal dependencies
    pubsub = test_import("pubsub")
    meta = test_import("meta")
    rsvg = test_import(
        "gi.repository.Rsvg", "SVG Support unavailable", gi_require=("Rsvg", "2.0")
    )

    return (
        test_import("shoebot"),
        AvailableModules(
            gi=gi, pgi=pgi, meta=meta, pubsub=pubsub, rsvg=rsvg, vext=vext, pango=pango
        ),
    )


def shoebot_example(**shoebot_kwargs):
    """
    Decorator to run some code in a bot instance.
    """

    def decorator(f):
        def run():
            from shoebot.util import ShoebotInstallError

            print("    Shoebot - %s:" % f.__name__.replace("_", " "))
            try:
                import shoebot

                outputfile = "/tmp/shoebot-%s.png" % f.__name__
                bot = shoebot.create_bot(outputfile=outputfile)
                f(bot)
                bot.finish()
                print("        [passed] : %s" % outputfile)
                print("")
            except ShoebotInstallError as e:
                print("        [failed]", e.args[0])
                print("")
            except Exception:
                print("        [failed] - traceback:")
                for line in traceback.format_exc().splitlines():
                    print("    %s" % line)
                print("")

        return run

    return decorator


@shoebot_example()
def standard_module_example(bot):
    bot.size(640, 480)
    bot.fill(1, 0.5, 0.1)
    bot.rect(10, 10, 100, 100)


@shoebot_example()
def module_using_text(bot):
    bot.size(640, 480)
    bot.stroke(0)
    bot.text("Should work with gi not pgi", 0, 0)


def display_graphics_implementation():
    print("Graphics Implementation:")
    try:
        from shoebot.core.backend import driver

        for k, v in list(driver.get_libs().items()):
            print("    %s: %s" % (k, v))
    except Exception as e:
        raise


def diagnose():
    display_platform()
    shoebot_module, available_modules = test_imports()
    if not shoebot_module:
        print("Skipping shoebot module tests.")
        return

    display_graphics_implementation()
    try:
        import shoebot
    except ImportError as e:
        print("Cannot 'import shoebot'")
        traceback.print_exc()
        return False

    print("\nShoebot Tests:")

    # shoebot itself
    standard_module_example()

    # shoebot with text (will fail under pypy or pgi)
    if available_modules.pango:
        module_using_text()


if __name__ == "__main__":
    diagnose()



import array
import os.path
from io import StringIO

from shoebot.core.backend import cairo, driver, gi
from shoebot.util import _copy_attrs

from .basecolor import ColorMixin
from .grob import Grob

try:
    gi.require_version("Rsvg", "2.0")
    from gi.repository import Rsvg
except ValueError:
    Rsvg = None


CENTER = "center"
CORNER = "corner"


class SurfaceRef(object):
    """ Cannot have a weakref to a cairo surface, so wrapper is used """

    def __init__(self, surface):
        self.surface = surface


class Image(Grob, ColorMixin):
    _surface_cache = (
        {}
    )  # Did have a WeakValueDictionary here but this caused a memory leak of images every frame
    _state_attributes = {
        "transform",
    }  # NBX uses transform and transformmode here

    def __init__(
        self,
        bot,
        path=None,
        x=0,
        y=0,
        width=None,
        height=None,
        alpha=1.0,
        data=None,
        **kwargs
    ):
        Grob.__init__(self, bot)
        ColorMixin.__init__(self, **kwargs)

        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.alpha = alpha
        self.path = path
        self.data = data
        sh = sw = None  # Surface Height and Width

        if isinstance(self.data, cairo.ImageSurface):
            sw = self.data.get_width()
            sh = self.data.get_height()
            self._surface = self.data
        else:
            # checks if image data is passed in command call, in this case it wraps
            # the data in a StringIO oject in order to use it as a file
            # the data itself must contain an entire image, not just pixel data
            # it can be useful for example to retrieve images from the web without
            # writing temp files (e.g. using nodebox's web library, see example 1 of the library)
            # if no data is passed the path is used to open a local file
            if self.data is None:
                surfaceref = self._surface_cache.get(path)
                if surfaceref:
                    surface = surfaceref.surface
                    if isinstance(surface, cairo.RecordingSurface):
                        extents = surface.get_extents()
                        # extents has x, y which we dont use right now
                        sw = extents.width
                        sh = extents.height
                    else:
                        sw = surface.get_width()
                        sh = surface.get_height()
                elif os.path.splitext(path)[1].lower() == ".svg" and Rsvg is not None:
                    handle = Rsvg.Handle()
                    svg = handle.new_from_file(path)
                    dimensions = svg.get_dimensions()
                    sw = dimensions.width
                    sh = dimensions.height
                    surface = cairo.RecordingSurface(
                        cairo.CONTENT_COLOR_ALPHA, (0, 0, sw, sh)
                    )
                    ctx = cairo.Context(surface)
                    pycairo_ctx = driver.ensure_pycairo_context(ctx)
                    svg.render_cairo(pycairo_ctx)
                elif os.path.splitext(path)[1].lower() == ".png":
                    surface = cairo.ImageSurface.create_from_png(path)
                    sw = surface.get_width()
                    sh = surface.get_height()
                else:
                    from PIL import Image as PILImage

                    with PILImage.open(path) as img:
                        if img.mode != "RGBA":
                            img = img.convert("RGBA")

                        sw, sh = img.size
                        # Would be nice to not have to do some of these conversions :-\
                        bgra_data = img.tobytes("raw", "BGRA", 0, 1)
                        bgra_array = array.array("B", bgra_data)
                        surface = cairo.ImageSurface.create_for_data(
                            bgra_array, cairo.FORMAT_ARGB32, sw, sh, sw * 4
                        )

                self._surface_cache[path] = SurfaceRef(surface)
            else:
                from PIL import Image as PILImage

                with PILImage.open(StringIO(self.data)) as img:
                    if img.mode != "RGBA":
                        img = img.convert("RGBA")

                    sw, sh = img.size
                    # Would be nice to not have to do some of these conversions :-\
                    bgra_data = img.tobytes("raw", "BGRA", 0, 1)
                    bgra_array = array.array("B", bgra_data)
                    surface = cairo.ImageSurface.create_for_data(
                        bgra_array, cairo.FORMAT_ARGB32, sw, sh, sw * 4
                    )

            if width is not None or height is not None:
                if width:
                    wscale = float(width) / sw
                else:
                    wscale = 1.0
                if height:
                    hscale = float(height) / sh
                else:
                    if width:
                        hscale = wscale
                    else:
                        hscale = 1.0
                self._transform.scale(wscale, hscale)

            self.width = width or sw
            self.height = height or sh
            self._surface = surface

        self._deferred_render()

    def _render(self, ctx):
        if self.width and self.height:
            # Go to initial point (CORNER or CENTER):
            transform = self._call_transform_mode(self._transform)

            ctx.set_matrix(transform)
            ctx.translate(self.x, self.y)
            ctx.set_source_surface(self._surface)
            ctx.paint()

    def draw(self):
        self._deferred_render()

    def _get_center(self):
        """Returns the center point of the path, disregarding transforms."""
        x = self.x + self.width / 2
        y = self.y + self.height / 2
        return x, y

    center = property(_get_center)

    def copy(self):
        p = self.__class__(
            self._bot, self.path, self.x, self.y, self.width, self.height
        )
        _copy_attrs(self._bot, p, self.state_attributes)
        return p



import os
from math import radians

from pkg_resources import resource_filename, Requirement

from shoebot.core.backend import cairo, gi, driver
from shoebot.core.events import REDRAW_EVENT, publish_event
from shoebot.sbio.socket_server import SocketServer

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

pycairo = driver.cairo

ICON_FILE = resource_filename(
    Requirement.parse("shoebot"), "share/pixmaps/shoebot-ide.png"
)


class BackingStore:
    instance = None

    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.surface = pycairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)

    @property
    def dimensions(self):
        return self.width, self.height

    @classmethod
    def get_backingstore(cls, width, height):
        if cls.instance is None or (width, height) != cls.instance.dimensions:
            cls.instance = BackingStore(width, height)
        return cls.instance


class ShoebotWidget(Gtk.DrawingArea, SocketServer):
    """
    Create a double buffered GTK+ widget on which we will draw using Cairo
    """

    # Draw in response to an expose-event
    def __init__(self, scale_fit=True, input_device=None):
        Gtk.DrawingArea.__init__(self)
        self.connect("size_allocate", self.on_resize)
        self.connect("draw", self.draw)

        self.scale_fit = scale_fit
        self.input_device = input_device

        self.width, self.height = 1, 1
        self.bot_size = None
        self.frame = -1
        self.backing_store = BackingStore.get_backingstore(1, 1)

    def draw_default_image(self, cr):
        if os.path.isfile(ICON_FILE):
            icon = cairo.ImageSurface.create_from_png(ICON_FILE)
            cr.set_source_surface(icon)
            cr.paint()
        else:
            # A hint that shoebot is running in developer mode.
            # Can probably be removed once bug 212 is fixed.
            # https://github.com/shoebot/shoebot/issues/212
            cr.push_group()
            cr.translate(48, 48)
            cr.rotate(radians(45.0))
            cr.rectangle(-32, -32, 64, 64)
            cr.pop_group()
            cr.set_source_rgb(1, 1, 0)
            cr.fill()

    def on_resize(self, widget, dimensions):
        self.width = dimensions.width
        self.height = dimensions.height
        if self.bot_size is not None:
            publish_event(REDRAW_EVENT, data=(self.width, self.height))

    def scale_context_and_center(self, cr):
        """
        Scale context based on difference between bot size and widget
        """
        bot_width, bot_height = self.bot_size
        if self.width != bot_width or self.height != bot_height:
            # Scale up by largest dimension
            if self.width < self.height:
                scale_x = float(self.width) / float(bot_width)
                scale_y = scale_x
                offset_x = 0.0
                offset_y = (self.height - (bot_height * scale_y)) / 2.0
                cr.translate(0, offset_y)
            elif self.width > self.height:
                scale_y = float(self.height) / float(bot_height)
                scale_x = scale_y
                offset_x = (self.width - (bot_width * scale_x)) / 2.0
                offset_y = 0.0
                cr.translate(offset_x, 0)
            else:
                scale_x = 1.0
                scale_y = 1.0
                offset_x = 0.0
                offset_y = 0.0
            cr.scale(scale_x, scale_y)
            self.input_device.offset_x = offset_x
            self.input_device.offset_y = offset_y
            self.input_device.scale_x = scale_y
            self.input_device.scale_y = scale_y

    def draw(self, widget, cr):
        """
        Draw just the exposed part of the backing store, scaled to fit
        """
        if self.bot_size is None:
            # No bot to draw yet.
            self.draw_default_image(cr)
            return

        cr = driver.ensure_pycairo_context(cr)

        surface = self.backing_store.surface
        cr.set_source_surface(surface)
        cr.paint()

    def create_rcontext(self, size, frame):
        """
        Creates a recording surface for the bot to draw on

        :param size: The width and height of bot
        """
        self.frame = frame
        width, height = size
        meta_surface = cairo.RecordingSurface(
            cairo.CONTENT_COLOR_ALPHA, (0, 0, width, height)
        )

        ctx = cairo.Context(meta_surface)
        self._ctx = ctx
        return ctx

    def do_drawing(self, size, frame, cairo_ctx):
        """
        Update the backing store from a cairo context and
        schedule a REDRAW_EVENT (expose event)

        :param size: width, height in pixels of bot
        :param frame: frame # thar was drawn
        :param cairo_ctx: cairo context the bot was drawn on
        """
        if self.get_window() and not self.bot_size:
            # Get initial size for window
            self.set_size_request(*size)

        self.bot_size = size
        self.backing_store = BackingStore.get_backingstore(self.width, self.height)

        cr = pycairo.Context(self.backing_store.surface)
        if self.scale_fit:
            self.scale_context_and_center(cr)

        cairo_ctx = driver.ensure_pycairo_context(cairo_ctx)
        cr.set_source_surface(cairo_ctx.get_target())
        # Create the cairo context
        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()
        self.queue_draw()



import os
import sys

from pathlib import Path
from shoebot.core.backend import gi
from shoebot.core.events import (
    publish_event,
    QUIT_EVENT,
    VARIABLE_CHANGED_EVENT,
    REDRAW_EVENT,
)
from gi.repository import Gdk, Gtk

from pkg_resources import resource_filename, Requirement

from shoebot.gui import ShoebotWidget, VarWindow
from shoebot.core import DrawQueueSink
from .gtk_input_device import GtkInputDeviceMixin

import locale
import gettext

APP = "shoebot"
DIR = sys.prefix + "/share/shoebot/locale"
ICON_FILE = resource_filename(
    Requirement.parse("shoebot"), "share/pixmaps/shoebot-ide.png"
)
locale.setlocale(locale.LC_ALL, "")
gettext.bindtextdomain(APP, DIR)
gettext.textdomain(APP)
_ = gettext.gettext


class ShoebotWindow(Gtk.Window, GtkInputDeviceMixin, DrawQueueSink):
    """
    Create a GTK+ window that contains a ShoebotWidget
    """

    # Draw in response to an expose-event
    def __init__(
        self,
        title=None,
        show_vars=False,
        menu_enabled=True,
        server=False,
        port=7777,
        fullscreen=False,
        outputfile=None,
    ):
        Gtk.Window.__init__(self)
        DrawQueueSink.__init__(self)
        GtkInputDeviceMixin.__init__(self)

        if os.path.isfile(ICON_FILE):
            self.set_icon_from_file(ICON_FILE)

        self.menu_enabled = menu_enabled
        self.has_server = server
        self.serverport = port
        self.show_vars = show_vars
        self.var_window = None
        self.is_fullscreen = False

        sb_widget = ShoebotWidget(input_device=self)

        self.title = title
        if title:
            self.set_title(title)
        if fullscreen:
            self.is_fullscreen = True
            self.fullscreen()
        self.connect("delete-event", self.do_window_close)

        self.sb_widget = sb_widget
        self.attach_gtk(self)

        self.uimanager = Gtk.UIManager()
        accelgroup = self.uimanager.get_accel_group()
        self.add_accel_group(accelgroup)

        self.action_group = Gtk.ActionGroup(name="Canvas")

        self.action_group.add_actions(
            [
                ("Save as", None, _("_Save as")),
                (
                    "svg",
                    "Save as SVG",
                    _("Save as _SVG"),
                    "<Control>1",
                    None,
                    self.snapshot_svg,
                ),
                (
                    "pdf",
                    "Save as PDF",
                    _("Save as _PDF"),
                    "<Control>2",
                    None,
                    self.snapshot_pdf,
                ),
                (
                    "ps",
                    "Save as PS",
                    _("Save as P_S"),
                    "<Control>3",
                    None,
                    self.snapshot_ps,
                ),
                (
                    "png",
                    "Save as PNG",
                    _("Save as P_NG"),
                    "<Control>4",
                    None,
                    self.snapshot_png,
                ),
                (
                    "close",
                    "Close window",
                    _("_Close Window"),
                    "<Control>w",
                    None,
                    self.do_window_close,
                ),
            ]
        )

        self.action_group.add_toggle_actions(
            [
                (
                    "vars",
                    "Variables Window",
                    _("Va_riables Window"),
                    "<Control>r",
                    None,
                    self.do_toggle_variables,
                    self.show_vars,
                ),
                (
                    "fullscreen",
                    "Fullscreen",
                    _("_Fullscreen"),
                    "<Control>f",
                    None,
                    self.do_toggle_fullscreen,
                    False,
                ),
                ("play", "Play", _("_Play"), "<Alt>p", None, self.do_toggle_play, True),
            ]
        )

        menuxml = """
        <popup action="Save as">
            <menuitem action="play"/>
            <menuitem action="vars"/>
            <menuitem action="fullscreen"/>
            <separator/>
            <menuitem action="svg"/>
            <menuitem action="ps"/>
            <menuitem action="pdf"/>
            <menuitem action="png"/>
            <separator/>
            <separator/>
            <menuitem action="close"/>
        </popup>
        """

        self.uimanager.insert_action_group(self.action_group, 0)
        self.uimanager.add_ui_from_string(menuxml)

        sb_widget.show()
        self.add(sb_widget)

        self.present()

        self.pending_snapshots = []  # list of filenames to save after rendering.
        if outputfile:
            # The test harness uses this to get the gui to output an image, otherwise it may not make much sense
            # to have an output file with a GUI window ?
            self.pending_snapshots.append(outputfile)

        while Gtk.events_pending():
            Gtk.main_iteration()

        self.window_open = True
        self.pause_speed = None  # TODO - factor out bot controller stuff

    def gtk_mouse_button_down(self, widget, event):
        """ Handle right mouse button clicks """
        if self.menu_enabled and event.button == 3:
            menu = self.uimanager.get_widget("/Save as")
            menu.popup(None, None, None, None, event.button, event.time)
        else:
            super(ShoebotWindow, self).gtk_mouse_button_down(widget, event)

    def render(self, size, frame, drawqueue):
        cairo_ctx = super(self.__class__, self).render(size, frame, drawqueue)
        self.sb_widget.do_drawing(size, frame, cairo_ctx)

    def rendering_finished(self, size, frame, r_context):
        """
        Save any snapshots that were scheduled from the GUI

        :param size:  width, height
        :param frame:   frame  number
        :param r_context:  cairo context
        """
        canvas = self.bot.canvas

        pending_snapshots = self.pending_snapshots
        for filename in pending_snapshots:
            # TODO, show filename in the GUI.
            print(f"Save snapshot: {filename}", file=sys.stderr)
            # TODO - remove many of these closures.
            f = canvas.output_closure(filename)
            f(r_context)

        super().rendering_finished(size, frame, r_context)
        self.pending_snapshots = []

    def create_rcontext(self, size, frame):
        """
        Delegates to the sb_widget
        """
        return self.sb_widget.create_rcontext(size, frame)

    def show_variables_window(self):
        """
        Show the variables window.
        """
        if self.var_window is None and self.bot._vars:
            self.var_window = VarWindow(
                self, self.bot, "%s variables" % (self.title or "Shoebot")
            )
            self.var_window.window.connect("destroy", self.var_window_closed)

    def hide_variables_window(self):
        """
        Hide the variables window
        """
        if self.var_window is not None:
            self.var_window.window.destroy()
            self.var_window = None

    def var_window_closed(self, widget):
        """
        Called if user clicked close button on var window
        :param widget:
        :return:
        """
        # TODO - Clean up the menu handling stuff its a bit spaghetti right now
        self.action_group.get_action("vars").set_active(False)
        self.show_vars = False
        self.var_window = None

    def var_changed(self, name, value):
        self.bot._namespace[name] = value
        if self.var_window:
            return self.var_window.update_var(name, value)
        else:
            v = self.bot._vars[name]
            publish_event(VARIABLE_CHANGED_EVENT, v)
            return True, value

    def output_image_filename(self, format):
        """
        :param format:  Format, e.g. svg, pdf, png
        :return:  Full image filename, based on bot name
        """
        script = self.bot._namespace["__file__"]
        frame = self.bot._namespace["ITERATION"]
        if script:
            return f"{Path(script).stem}-{frame:04}.{format}"

        return f"output.{format}"

    def snapshot_svg(self, widget):
        """
        Request to save an SVG file after drawing is complete.
        """
        self.pending_snapshots.append(self.output_image_filename("svg"))
        publish_event(
            REDRAW_EVENT, data=(None, None)
        )  # TODO - this probably wants its own event

    def snapshot_ps(self, widget):
        """
        Request to save a Postscript file after drawing is complete.
        """
        self.pending_snapshots.append(self.output_image_filename("ps"))
        publish_event(
            REDRAW_EVENT, data=(None, None)
        )  # TODO - this probably wants its own event

    def snapshot_pdf(self, widget):
        """
        Request to save a PDF file after drawing is complete.
        """
        self.pending_snapshots.append(self.output_image_filename("pdf"))
        publish_event(
            REDRAW_EVENT, data=(None, None)
        )  # TODO - this probably wants its own event

    def snapshot_png(self, widget):
        """
        Request to save a PNG file after drawing is complete.
        """
        self.pending_snapshots.append(self.output_image_filename("png"))
        publish_event(
            REDRAW_EVENT, data=(None, None)
        )  # TODO - this probably wants its own event

    def trigger_fullscreen_action(self, fullscreen):
        """
        Toggle fullscreen from outside the GUI,
        causes the GUI to updated and run all its actions.
        """
        action = self.action_group.get_action("fullscreen")
        action.set_active(fullscreen)

    def do_fullscreen(self, widget):
        """
        Widget Action to Make the window fullscreen and update the bot.
        """
        self.fullscreen()
        self.is_fullscreen = True
        # next lines seem to be needed for window switching really to
        # fullscreen mode before reading it's size values
        while Gtk.events_pending():
            Gtk.main_iteration()
        # we pass the full-screen size to bot
        self.bot._screen_width = Gdk.Screen.width()
        self.bot._screen_height = Gdk.Screen.height()
        self.bot._screen_ratio = self.bot._screen_width / self.bot._screen_height

    def do_unfullscreen(self, widget):
        """
        Widget Action to set Windowed Mode.
        """
        self.unfullscreen()
        self.is_fullscreen = False
        self.bot._screen_ratio = None

    def do_window_close(self, widget, data=None):
        """
        Widget Action to Close the window, triggering the quit event.
        """
        publish_event(QUIT_EVENT)

        if self.has_server:
            self.sock.close()

        self.hide_variables_window()

        self.destroy()
        self.window_open = False

    def do_toggle_fullscreen(self, action):
        """
        Widget Action to Toggle fullscreen from the GUI
        """
        is_fullscreen = action.get_active()
        if is_fullscreen:
            self.fullscreen()
        else:
            self.unfullscreen()

    def do_toggle_play(self, action):
        """
        Widget Action to toggle play / pause.
        """
        # TODO - move this into bot controller
        # along with stuff in socketserver and shell
        if self.pause_speed is None and not action.get_active():
            self.pause_speed = self.bot._speed
            self.bot._speed = 0
        else:
            self.bot._speed = self.pause_speed
            self.pause_speed = None

    def do_toggle_variables(self, action):
        """
        Widget Action to toggle showing the variables window.
        """
        self.show_vars = action.get_active()
        if self.show_vars:
            self.show_variables_window()
        else:
            self.hide_variables_window()

    def main_iteration(self):
        """
        Called from main loop, if your sink needs to handle GUI events
        do it here.

        Check any GUI flags then call Gtk.main_iteration to update things.
        """
        if self.show_vars:
            self.show_variables_window()
        else:
            self.hide_variables_window()

        while Gtk.events_pending():
            Gtk.main_iteration()



import os.path
import sys

from shoebot.core.backend import cairo
from shoebot.kgp import KantGenerator
from shoebot.data import ShoebotError
from shoebot.util.fonts import list_pango_fonts
from .bot import Bot
from shoebot.data import (
    geometry,
    Point,
    BezierPath,
    Image,
    RGB,
    HSB,
    CORNER,
    CENTER,
    MOVETO,
    RMOVETO,
    LINETO,
    RLINETO,
    CURVETO,
    RCURVETO,
    ARC,
    ELLIPSE,
    CLOSE,
    LEFT,
    RIGHT,
    BUTT,
    ROUND,
    SQUARE,
    BEVEL,
    MITER,
)

from math import sin, cos, pi
from math import radians as deg2rad

import locale
import gettext

APP = "shoebot"
DIR = sys.prefix + "/share/shoebot/locale"
locale.setlocale(locale.LC_ALL, "")
ASSETS_DIR = sys.prefix + "/share/shoebot/data"
gettext.bindtextdomain(APP, DIR)
# gettext.bindtextdomain(APP)
gettext.textdomain(APP)
_ = gettext.gettext

# Nodebox compatibility shim
sys.path.append(os.path.join(os.path.dirname(__file__), "nodebox-lib"))
sys.path.append(".")  # ximport can work from current dir


class NodeBot(Bot):

    NORMAL = "1"
    FORTYFIVE = "2"

    PIE = "pie"
    CHORD = "chord"

    CORNER = CORNER
    CENTER = CENTER
    MOVETO = MOVETO
    RMOVETO = RMOVETO
    LINETO = LINETO
    RLINETO = RLINETO
    CURVETO = CURVETO
    RCURVETO = RCURVETO
    ARC = ARC
    ELLIPSE = ELLIPSE  # Shoebot, not nodebox 1.x
    CLOSE = CLOSE

    LEFT = LEFT
    RIGHT = RIGHT

    # Default values
    color_mode = RGB
    color_range = 1

    def __init__(self, canvas=None, namespace=None, vars=None):
        """
        Nodebot grammar constructor

        :param canvas: Canvas implementation for output.
        :param namespace: Optionally specify a dict to inject as namespace
        :param vars: Optional dict containing initial values for variables
        """
        Bot.__init__(self, canvas, namespace=namespace, vars=vars)
        canvas.mode = CORNER

    @property
    def _ns(self):
        """Nodebox1 API way of fetching namespace from _ctx"""
        return self._namespace

    # Drawing

    # Image

    def image(
        self,
        path,
        x,
        y,
        width=None,
        height=None,
        alpha=1.0,
        data=None,
        draw=True,
        **kwargs,
    ):
        """Draws a image with (x,y) as the top left corner.

        If width and height are specified, resize the image to fit.

        form path, in x,y and resize it to width, height dimensions.

        :param path: location of the image on disk
        :param x: x-coordinate of the top left corner
        :param y: y-coordinate of the top left corner
        :param width: image width (leave blank to use its original width)
        :param height: image height (leave blank to use its original height)
        :param alpha: opacity
        :param data: image data to load. Use this instead of ``path`` if you want to load an image from memory or have another source (e.g. using the `web` library)
        :param draw: whether to place the image immediately on the canvas or not
        :type path: filename
        :type x: float
        :type y: float
        :type width: float or None
        :type height: float or None
        :type alpha: float
        :type data: binary data
        :type draw: bool
        """

        return self.Image(path, x, y, width, height, alpha, data, **kwargs)

    def imagesize(self, path):
        """
        :param path: Path to image file.
        :return: image size as a (width, height) tuple
        """
        from PIL import Image as PILImage

        with PILImage.open(path) as img:
            return img.size

    # Paths

    def rect(self, x, y, width, height, roundness=0.0, draw=True, **kwargs):
        """Draw a rectangle.

        :param x: top left x-coordinate
        :param y: top left y-coordinate
        :param width: rectangle width
        :param height: rectangle height
        :param roundness: rounded corner radius
        :param boolean draw: whether to draw the shape on the canvas or not
        :param fill: fill color
        :return: BezierPath representing the rectangle
        """
        path = self.BezierPath(**kwargs)
        path.rect(x, y, width, height, roundness, self.rectmode)
        if draw:
            path.draw()
        return path

    def rectmode(self, mode=None):
        """Get or set the current rectmode.

        :param mode: the mode to draw new rectangles in
        :type mode: CORNER, CENTER or CORNERS
        :return: current rectmode value
        """
        if mode in (self.CORNER, self.CENTER, self.CORNERS):
            self.rectmode = mode
            return self.rectmode
        elif mode is None:
            return self.rectmode
        else:
            raise ShoebotError(_("rectmode: invalid input"))

    def ellipse(self, x, y, width, height, draw=True, **kwargs):
        """Draw an ellipse.

        :param x: top left x-coordinate
        :param y: top left y-coordinate
        :param width: ellipse width
        :param height: ellipse height
        :param boolean draw: whether to draw the shape on the canvas or not
        :return: BezierPath representing the ellipse
        """

        path = self.BezierPath(**kwargs)
        path.ellipse(x, y, width, height, self.ellipsemode)
        if draw:
            path.draw()
        return path

    oval = ellipse

    def circle(self, x, y, diameter, draw=True, **kwargs):
        """Draw a circle
        :param x: x-coordinate of the top left corner
        :param y: y-coordinate of the top left corner
        :param diameter: circle diameter
        :param boolean draw: whether to draw the shape on the canvas or not
        :return: BezierPath representing the circle
        """
        return self.ellipse(x, y, diameter, diameter, draw, **kwargs)

    def ellipsemode(self, mode=None):
        """
        Set the current ellipse drawing mode.

        :param mode: CORNER, CENTER, CORNERS
        :return: ellipsemode if mode is None or valid.
        """
        if mode in (self.CORNER, self.CENTER, self.CORNERS):
            self.ellipsemode = mode
            return self.ellipsemode
        elif mode is None:
            return self.ellipsemode
        else:
            raise ShoebotError(_("ellipsemode: invalid input"))

    def line(self, x1, y1, x2, y2, draw=True, **kwargs):
        """Draw a line from (x1,y1) to (x2,y2).

        :param x1: x-coordinate of the first point
        :param y1: y-coordinate of the first point
        :param x2: x-coordinate of the second point
        :param y2: y-coordinate of the second point
        :param boolean draw: whether to draw the shape on the canvas or not
        :return: BezierPath representing the line
        """
        self.beginpath(**kwargs)
        self.moveto(x1, y1)
        self.lineto(x2, y2)
        return self.endpath(draw=draw, closed=False)

    def arc(self, x, y, radius, angle1, angle2, type=CHORD, draw=True, **kwargs):
        """Draw an arc with center (x,y) between two angles in degrees.

        :param x1: start x-coordinate
        :param y1: start y-coordinate
        :param radius: arc radius
        :param angle1: start angle
        :param angle2: end angle
        """
        self.beginpath(**kwargs)
        if type == self.PIE:
            # find the coordinates of the start and end points
            x1 = x + radius * cos(deg2rad(angle1))
            y1 = y + radius * sin(deg2rad(angle1))
            x2 = x + radius * cos(deg2rad(angle2))
            y2 = y + radius * sin(deg2rad(angle2))
            self.moveto(x2, y2)
            self.lineto(x, y)
            self.lineto(x1, y1)
        # self.moveto(x, y)  # uncomment to fix the "empty path" issue
        self.arcto(x, y, radius, angle1, angle2)
        return self.endpath(draw=draw)

    def arrow(self, x, y, width, type=NORMAL, draw=True, **kwargs):
        """Draw an arrow.

        :param x: arrow tip x-coordinate
        :param y: arrow tip y-coordinate
        :param width: arrow width (also sets height)
        :param type: arrow type
        :type type: NORMAL or FORTYFIVE
        :param boolean draw: whether to draw the shape on the canvas or not
        :return: BezierPath object representing the arrow
        """
        # Taken from Nodebox
        self.beginpath(**kwargs)
        if type == self.NORMAL:
            head = width * 0.4
            tail = width * 0.2
            self.moveto(x, y)
            self.lineto(x - head, y + head)
            self.lineto(x - head, y + tail)
            self.lineto(x - width, y + tail)
            self.lineto(x - width, y - tail)
            self.lineto(x - head, y - tail)
            self.lineto(x - head, y - head)
            self.lineto(x, y)
        elif type == self.FORTYFIVE:
            head = 0.3
            tail = 1 + head
            self.moveto(x, y)
            self.lineto(x, y + width * (1 - head))
            self.lineto(x - width * head, y + width)
            self.lineto(x - width * head, y + width * tail * 0.4)
            self.lineto(x - width * tail * 0.6, y + width)
            self.lineto(x - width, y + width * tail * 0.6)
            self.lineto(x - width * tail * 0.4, y + width * head)
            self.lineto(x - width, y + width * head)
            self.lineto(x - width * (1 - head), y)
            self.lineto(x, y)
        else:
            raise NameError(
                _("arrow: available types for arrow() are NORMAL and FORTYFIVE\n")
            )
        return self.endpath(draw=draw)

    def star(self, x, y, points=20, outer=100, inner=50, draw=True, **kwargs):
        """Draws a star.

        :param x: center x-coordinate
        :param y: center y-coordinate
        :param points: amount of points
        :param outer: outer radius
        :param inner: inner radius
        :param boolean draw: whether to draw the shape on the canvas or not
        """

        # Taken from Nodebox.
        self.beginpath(**kwargs)
        self.moveto(x, y + outer)

        for i in range(1, int(2 * points)):
            angle = i * pi / points
            px = sin(angle)
            py = cos(angle)
            if i % 2:
                radius = inner
            else:
                radius = outer
            px = x + radius * px
            py = y + radius * py
            self.lineto(px, py)

        return self.endpath(draw)

    # Path
    # Path functions taken from Nodebox and modified

    def beginpath(self, x=None, y=None, **kwargs):
        """Start a new Bézier path.

        This command is needed before any other path drawing commands.

        :param x: x-coordinate of the starting point
        :param y: y-coordinate of the starting point
        :type x: float or None
        :type y: float or None
        """
        self._path = self.BezierPath(**kwargs)
        # if we have arguments, do a moveto too
        if x is not None and y is not None:
            self._path.moveto(x, y)

    def moveto(self, x, y):
        """Move the Bézier "pen" to the specified point without drawing.

        :param x: x-coordinate of the point to move to
        :param y: y-coordinate of the point to move to
        :type x: float
        :type y: float
        """
        if self._path is None:
            # self.beginpath()
            raise ShoebotError(_("No current path. Use beginpath() first."))
        self._path.moveto(x, y)

    def lineto(self, x, y):
        """Draw a line from the pen's current point.

        :param x: x-coordinate of the point to draw to
        :param y: y-coordinate of the point to draw to
        :type x: float
        :type y: float
        """
        if self._path is None:
            raise ShoebotError(_("No current path. Use beginpath() first."))
        self._path.lineto(x, y)

    def curveto(self, x1, y1, x2, y2, x3, y3):
        if self._path is None:
            raise ShoebotError(_("No current path. Use beginpath() first."))
        self._path.curveto(x1, y1, x2, y2, x3, y3)

    def arcto(self, x, y, radius, angle1, angle2):
        if self._path is None:
            raise ShoebotError(_("No current path. Use beginpath() first."))
        # use degrees by default
        angle1 = deg2rad(angle1)
        angle2 = deg2rad(angle2)
        self._path.arc(x, y, radius, angle1, angle2)

    def closepath(self):
        if self._path is None:
            raise ShoebotError(_("No current path. Use beginpath() first."))
        if not self._path.closed:
            self._path.closepath()
            self._path.closed = True

    def endpath(self, draw=True, closed=None):
        if self._path is None:
            raise ShoebotError(_("No current path. Use beginpath() first."))
        p = self._path
        if self._autoclosepath is True and closed is not False:
            self._path.closepath()
        if draw:
            p.draw()
        else:
            # keep the transform so we don't lose it
            self._path.transform = cairo.Matrix(*self._canvas.transform)
        self._path = None
        return p

    def drawpath(self, path, **kwargs):
        if isinstance(path, BezierPath):
            p = self.BezierPath(path=path, **kwargs)
            p.draw()
        elif isinstance(path, Image):
            path.draw()  # Is this right ? - added to make test_clip_4.bot work
        elif hasattr(path, "__iter__"):
            p = self.BezierPath()
            for point in path:
                p.addpoint(point)
            p.draw()

    def drawimage(self, image, x=None, y=None):
        """
        :param image: Image to draw
        :param x: optional, x coordinate (default is image.x)
        :param y: optional, y coordinate (default is image.y)
        :return:
        """
        if x is None:
            x = image.x
        if y is None:
            y = image.y
        self.image(image.path, image.x, image.y, data=image.data)

    def autoclosepath(self, close=True):
        self._autoclosepath = close

    def relmoveto(self, x, y):
        """Move relatively to the last point."""
        if self._path is None:
            raise ShoebotError(_("No current path. Use beginpath() first."))
        self._path.relmoveto(x, y)

    def rellineto(self, x, y):
        """Draw a line using relative coordinates."""
        if self._path is None:
            raise ShoebotError(_("No current path. Use beginpath() first."))
        self._path.rellineto(x, y)

    def relcurveto(self, h1x, h1y, h2x, h2y, x, y):
        """Draws a curve relatively to the last point."""
        if self._path is None:
            raise ShoebotError(_("No current path. Use beginpath() first."))
        self._path.relcurveto(h1x, h1y, h2x, h2y, x, y)

    def findpath(self, points, curvature=1.0):

        """Constructs a path between the given list of points.

        Interpolates the list of points and determines
        a smooth bezier path betweem them.

        The curvature parameter offers some control on
        how separate segments are stitched together:
        from straight angles to smooth curves.
        Curvature is only useful if the path has more than  three points.
        """

        # The list of points consists of Point objects,
        # but it shouldn't crash on something straightforward
        # as someone supplying a list of (x,y)-tuples.

        for i, pt in enumerate(points):
            if type(pt) == tuple:
                points[i] = Point(pt[0], pt[1])

        if len(points) == 0:
            return None
        if len(points) == 1:
            path = self.BezierPath(None)
            path.moveto(points[0].x, points[0].y)
            return path
        if len(points) == 2:
            path = self.BezierPath(None)
            path.moveto(points[0].x, points[0].y)
            path.lineto(points[1].x, points[1].y)
            return path

        # Zero curvature means straight lines.

        curvature = max(0, min(1, curvature))
        if curvature == 0:
            path = self.BezierPath(None)
            path.moveto(points[0].x, points[0].y)
            for i in range(len(points)):
                path.lineto(points[i].x, points[i].y)
            return path

        curvature = 4 + (1.0 - curvature) * 40

        dx = {0: 0, len(points) - 1: 0}
        dy = {0: 0, len(points) - 1: 0}
        bi = {1: -0.25}
        ax = {1: (points[2].x - points[0].x - dx[0]) / 4}
        ay = {1: (points[2].y - points[0].y - dy[0]) / 4}

        for i in range(2, len(points) - 1):
            bi[i] = -1 / (curvature + bi[i - 1])
            ax[i] = -(points[i + 1].x - points[i - 1].x - ax[i - 1]) * bi[i]
            ay[i] = -(points[i + 1].y - points[i - 1].y - ay[i - 1]) * bi[i]

        r = list(range(1, len(points) - 1))
        r.reverse()
        for i in r:
            dx[i] = ax[i] + dx[i + 1] * bi[i]
            dy[i] = ay[i] + dy[i + 1] * bi[i]

        path = self.BezierPath(None)
        path.moveto(points[0].x, points[0].y)
        for i in range(len(points) - 1):
            path.curveto(
                points[i].x + dx[i],
                points[i].y + dy[i],
                points[i + 1].x - dx[i + 1],
                points[i + 1].y - dy[i + 1],
                points[i + 1].x,
                points[i + 1].y,
            )

        return path

    # Transform and utility

    def beginclip(self, path):
        """Use a path as a clipping mask.

        All drawing commands between beginclip() and endclip() will be drawn
        inside the clipping mask set by beginclip().

        :param path: the path to be used as a clipping mask
        :type path: BezierPath
        """
        p = self.ClippingPath(path)
        p.draw()
        return p

    def endclip(self):
        """Finish a clipping mask and render the result."""
        p = self.EndClip()
        p.draw()

    def transform(self, mode=None):
        """Set the current transform mode.

        :param mode: the mode to base new transformations on
        :type mode: CORNER or CENTER
        """
        if mode:
            self._canvas.mode = mode
        return self._canvas.mode

    def translate(self, xt, yt):
        """
        Translate the canvas origin point by (xt, yt).

        :param xt: Amount to move horizontally
        :param yt: Amount to move vertically
        """
        self._canvas.translate(xt, yt)

    def rotate(self, degrees=0, radians=0):
        """
        Set the current rotation in degrees or radians.

        :param degrees: Degrees to rotate
        :param radians: Radians to rotate
        """
        # TODO change canvas to use radians
        if radians:
            angle = radians
        else:
            angle = deg2rad(degrees)
        self._canvas.rotate(-angle)

    def scale(self, x=1, y=None):
        """
        Set a scale at which to draw objects.

        1.0 draws objects at their natural size.

        :param x: Scale on the horizontal plane
        :param y: Scale on the vertical plane
        """
        if not y:
            y = x
        if x == 0:
            # Cairo borks on zero values
            x = 1
        if y == 0:
            y = 1
        self._canvas.scale(x, y)

    def skew(self, x=1, y=0):
        # TODO bring back transform mixin
        t = self._canvas.transform
        t *= cairo.Matrix(1, 0, x, 1, 0, 0)
        t *= cairo.Matrix(1, y, 0, 1, 0, 0)
        self._canvas.transform = t

    def push(self):
        self._canvas.push_matrix()

    def pop(self):
        self._canvas.pop_matrix()

    def reset(self):
        self._canvas.reset_transform()

    # Color

    def outputmode(self):
        """
        NOT IMPLEMENTED
        """
        raise NotImplementedError(
            _("outputmode() isn't implemented; Shoebot does not support CMYK")
        )

    def colormode(self, mode=None, range=None):
        """Set the current colormode (can be RGB or HSB) and eventually
        the color range.

        If called without arguments, it returns the current colormode.

        :param mode: Color mode to use
        :type mode: RGB or HSB
        :param float crange: Maximum value for the new color range to use
        :return: Current color mode
        """
        if mode is not None:
            if mode == RGB:
                self.color_mode = RGB
            elif mode == HSB:
                self.color_mode = HSB
            else:
                raise NameError(_("Only RGB and HSB colormodes are supported."))
        if range is not None:
            self.color_range = range
        return self.color_mode

    def colorrange(self, crange=None):
        """Sets the current color range.

        The default is 0-1; for a range of 0-255, use ``colorrange(256)``.

        :param float crange: maximum value for new color range
        :return: current range value
        """
        if crange is not None:
            self.color_range = float(crange)
        return self.color_range

    def fill(self, *args):
        """Sets a fill color, applying it to new paths.

        :param args: color in supported format
        """
        if args is not None:
            self._canvas.fillcolor = self.color(*args)
        return self._canvas.fillcolor

    def nofill(self):
        """Stop applying fills to new paths.

        :return: fill color before nofill() was called
        """
        c = self._canvas.fillcolor
        self._canvas.fillcolor = None
        return c

    def fillrule(self, r=None):
        """Set the fill rule to use in new paths.

        :param r: fill rule to apply
        :type r: WINDING or EVENODD
        :return: current fill rule value
        """
        if r is not None:
            self._canvas.fillrule = r
        return self._canvas.fillrule

    def stroke(self, *args):
        """Set a stroke color, applying it to new paths.

        :param args: color in supported format
        :return: new stroke color
        """
        if args is not None:
            self._canvas.strokecolor = self.color(*args)
        return self._canvas.strokecolor

    def nostroke(self):
        """Stop applying strokes to new paths.

        :return: stroke color before nostroke() was called
        """
        c = self._canvas.strokecolor
        self._canvas.strokecolor = None
        return c

    def strokewidth(self, w=None):
        """Set the stroke width to be used by stroke().

        :param w: width of the stroke to use
        :return: current stroke width value
        """
        if w is not None:
            self._canvas.strokewidth = w
        return self._canvas.strokewidth

    def strokedash(self, dashes=None, offset=0):
        """Sets the dash pattern to be used by stroke().

        :param list dashes: a sequence specifying alternate lengths of on and off stroke portions
        :param float offset: an offset into the dash pattern at which the stroke should start
        :return: tuple with dashes value and offset
        """
        if dashes is not None:
            self._canvas.strokedash = dashes
        if offset:
            self._canvas.dashoffset = offset
        return (self._canvas.strokedash, self._canvas.dashoffset)

    def strokecap(self, cap=None):
        """Set the stroke cap.

        :param w: new stroke cap value
        :return: current stroke cap value
        """
        if cap is not None:
            self._canvas.strokecap = cap
        return self._canvas.strokecap

    def strokejoin(self, join=None):
        """Set the stroke join.

        :param w: new stroke join value
        :return: current line join value
        """
        if join is not None:
            self._canvas.strokejoin = join
        return self._canvas.strokejoin

    def background(self, *args):
        """Set the canvas background color.

        :param color: background color to apply
        :return: new background color
        """
        self._canvas.background = self.color(*args)
        return self._canvas.background

    def blendmode(self, mode=None):
        """Set the current blending mode.

        :param mode: mode name (e.g. "multiply")
        """
        if mode:
            self._canvas.blendmode = mode
        return self._canvas.blendmode

    # Text

    def font(self, fontpath=None, fontsize=None, vars=None, *args, **kwargs):
        """Set the font to be used with new text instances.

        :param fontpath: font name (can include styles like "Bold")
        :param fontsize: font size
        :param vars: font variant values, as a dict of axis/value pairs (variable fonts only)
        :param var_XXXX: set variant value (variable fonts only)
        :return: current fontpath (if fontpath param not set)

        Accepts TrueType and OpenType files. Depends on FreeType being
        installed."""
        if fontpath is not None:
            # do we have variants set?
            if not vars:
                # make a list of "arg=value" strings to append to the font name below
                variants = [
                    f"{arg.replace('var_', '')}={value}"
                    for arg, value in kwargs.items()
                    if arg.startswith("var_")
                ]
            else:
                # make a list of "arg=value" strings from the provided dict
                variants = [f"{arg}={value}" for arg, value in vars.items()]
            if variants:
                # append to the font string
                # syntax: "Inconsolata @wdth=50,wght=600"
                fontpath += " @" + ",".join(variants)
            self._canvas.fontfile = fontpath

        if fontsize is not None:
            self._canvas.fontsize = fontsize

        return self._canvas.fontfile

    def fontsize(self, fontsize=None):
        """Sets and/or returns the current font size.

        :param fontsize: Font size in pt
        :return: the current font size value
        """
        if fontsize is not None:
            self._canvas.fontsize = fontsize
        return self._canvas.fontsize

    def text(
        self, txt, x, y, width=None, height=1000000, outline=False, draw=True, **kwargs
    ):
        """Draws a string of text according to the current font settings.

        :param txt: Text to output
        :param x: x-coordinate of the top left corner
        :param y: y-coordinate of the top left corner
        :param width: text width
        :param height: text height
        :param outline: If True, draws a path instead of a text object (defaults to False)
        :param draw: Set to False to inhibit immediate drawing (defaults to True)
        :return: Path object representing the text.
        """
        txt = self.Text(txt, x, y, width, height, outline=outline, ctx=None, **kwargs)
        if outline:
            path = txt.path
            if draw:
                path.draw()
            return path
        return txt

    def textpath(self, txt, x, y, width=None, height=1000000, draw=False, **kwargs):
        """Generates an outlined path of the input text.

        :param txt: Text to output
        :param x: x-coordinate of the top left corner
        :param y: y-coordinate of the top left corner
        :param width: text width
        :param height: text height
        :param draw: Set to False to inhibit immediate drawing (defaults to False)
        :return: BezierPath representing the text
        """
        txt = self.Text(txt, x, y, width, height, draw=False, **kwargs)
        path = txt.path
        if draw:
            path.draw()
        return path

    def textmetrics(self, txt, width=None, height=None, **kwargs):
        """Returns the dimensions of the text box of a string of text, according
        to the current font settings.

        :return: (width, height) tuple
        """
        # for now only returns width and height (as per Nodebox behaviour)
        # but maybe we could use the other data from cairo
        txt = self.Text(txt, 0, 0, width, height, draw=False, **kwargs)
        return txt.metrics

    def textbounds(self, txt, width=None, height=None, **kwargs):
        """Returns the dimensions of the actual shapes (inked part) of a string
        of text, according to the current font settings.

        :return: (width, height) tuple
        """
        txt = self.Text(txt, 0, 0, width, height, draw=False, **kwargs)
        return txt.bounds

    def textwidth(self, txt, width=None, **kwargs):
        """Returns the width of a string of text according to the current
        font settings.

        :return:
        """
        w = width
        return self.textmetrics(txt, width=w, **kwargs)[0]

    def textheight(self, txt, width=None, **kwargs):
        """Returns the height of a string of text according to the current
        font settings.

        :param txt: string to measure
        :param width: width of a line of text in a block
        """
        w = width
        return self.textmetrics(txt, width=w, **kwargs)[1]

    def lineheight(self, height=None):
        """Set text lineheight.

        :param height: line height.
        """
        if height is not None:
            self._canvas.lineheight = height
        return self._canvas.lineheight

    def align(self, align=LEFT):
        """Set text alignment

        :param align: Text alignment (LEFT, CENTER, RIGHT)
        """
        self._canvas.align = align
        return self._canvas.align

    def fontoptions(
        self, hintstyle=None, hintmetrics=None, subpixelorder=None, antialias=None
    ):
        """Set font rendering options.

        :param hintstyle: Hinting style (NONE, SLIGHT, MEDIUM, FULL)
        :param hintmetrics: Quantize font metrics (ON, OFF)
        :param antialias: Antialiasing type (NONE, GRAY, SUBPIXEL, FAST, GOOD, BEST)
        :param subpixelorder: Order of pixels when antialiasing in SUBPIXEL mode (RGB, BGR, VRGB, VBGR)
        """
        self._canvas.hintstyle = hintstyle
        self._canvas.hintmetrics = hintmetrics
        self._canvas.subpixelorder = subpixelorder
        self._canvas.antialias = antialias

    def fontnames(self):
        return list_pango_fonts()

    def autotext(self, sourceFile):
        k = KantGenerator(sourceFile, searchpaths=[".", ASSETS_DIR])
        return k.output()

    @property
    def canvas(self):
        """
        Not entirely sure compatible the Shoebot 'canvas' is with Nodebox
        but there you go.
        :return:
        """
        return self._canvas

    def angle(self, x0, y0, x1, y1):
        return geometry.angle(x0, y0, x1, y1)

    def distance(self, x0, y0, x1, y1):
        return geometry.distance(x0, y0, x1, y1)

    def coordinates(self, x0, y0, distance, angle):
        return geometry.coordinates(x0, y0, distance, angle)



#!/usr/bin/env python3
import argparse
import colorsys
import errno
import gettext
import locale
import os
import sys

import shoebot

from shoebot.data import ShoebotError
from shoebot.core.backend import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
gi.require_version("GtkSource", "3.0")

from gi.repository import Gdk, GLib, GObject, Gtk, GtkSource, Pango

DEBUG = os.environ.get("SHOEBOT_DEBUG_IDE", "0").lower() in ["1", "yes", "true"]

APP = "shoebot"
LOCALE_DIR = sys.prefix + "/share/shoebot/locale"

SEARCH_BACKWARD = -1
SEARCH_FORWARD = 1

locale.setlocale(locale.LC_ALL, "")
gettext.bindtextdomain(APP, LOCALE_DIR)
gettext.textdomain(APP)
_ = gettext.gettext

ICON_FILE = f"{sys.prefix}/shoebot/icon.png"

GTK_WRAP_MODES = {
    "WrapWords": Gtk.WrapMode.WORD,
    "WrapChars": Gtk.WrapMode.CHAR,
    "WrapNone": Gtk.WrapMode.NONE,
}


def hue_to_rgb(hue):
    if hue > 1.0:
        raise ValueError("Hue cannot be > 1.0")

    return colorsys.hsv_to_rgb(hue, 1.0, 1.0)


class SourceBuffer(GtkSource.Buffer):
    N_COLORS = 16
    PANGO_SCALE = 1024

    def __init__(self, filename=None):
        GObject.GObject.__init__(self)

        self.refcount = 0
        if filename is None:
            self.filename = ShoebotIDE.get_next_untitled_filename()
        else:
            self.open_file(filename)
            self.filename = filename

        self.color_tags = [self.create_tag() for i in range(self.N_COLORS)]
        self.color_cycle_timeout_id = 0
        self.start_hue = 0.0

        # self.invisible_tag = self.create_tag(None, invisible=True)
        self.not_editable_tag = self.create_tag(editable=False, foreground="purple")
        self.found_text_tag = self.create_tag(foreground="red")

        tabs = Pango.TabArray.new(4, True)
        tabs.set_tab(0, Pango.TabAlign.LEFT, 10)
        tabs.set_tab(1, Pango.TabAlign.LEFT, 30)
        tabs.set_tab(2, Pango.TabAlign.LEFT, 60)
        tabs.set_tab(3, Pango.TabAlign.LEFT, 120)
        self.custom_tabs_tag = self.create_tag(tabs=tabs, foreground="green")

    def pretty_name(self):
        return os.path.basename(self.filename)

    def search(self, text, view, direction=SEARCH_FORWARD):
        if not text:
            return
        # remove tag from whole buffer
        start, end = self.get_bounds()
        self.remove_tag(self.found_text_tag, start, end)

        it = self.get_iter_at_mark(self.get_insert())

        total_matches = 0
        while True:
            if direction == SEARCH_FORWARD:
                match = it.forward_search(text, Gtk.TextSearchFlags.TEXT_ONLY)
            else:
                match = it.backward_search(text, Gtk.TextSearchFlags.TEXT_ONLY)
            if not match:
                break
            match_start, match_end = match
            total_matches += 1
            self.apply_tag(self.found_text_tag, match_start, match_end)
            it = match_end

        dialog = Gtk.MessageDialog(
            view,
            Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.INFO,
            Gtk.ButtonsType.OK,
            _("%d strings found and marked in red") % total_matches,
        )
        dialog.connect("response", lambda x, y: dialog.destroy())
        dialog.show()

    def ref(self):
        self.refcount += 1

    def unref(self):
        self.refcount -= 1
        if self.refcount == 0:
            self.set_colors(False)
            del self

    def color_cycle_timeout(self):
        self.cycle_colors()
        return True

    def set_colors(self, enabled):
        hue = 0.0

        if enabled and self.color_cycle_timeout_id == 0:
            self.color_cycle_timeout_id = GLib.timeout_add(
                200, self.color_cycle_timeout
            )
        elif not enabled and self.color_cycle_timeout_id:
            GLib.source_remove(self.color_cycle_timeout_id)
            self.color_cycle_timeout_id = 0

        for tag in self.color_tags:
            if enabled:
                color = Gdk.RGBA(*hue_to_rgb(hue))
                tag.set_property("foreground_rgba", color)
            else:
                tag.set_property("foreground_set", False)
            hue += 1.0 / SourceBuffer.N_COLORS

    def cycle_colors(self):
        hue = self.start_hue

        for tag in self.color_tags:
            color = Gdk.Color(*hue_to_rgb(hue))
            tag.set_property("foreground_gdk", color)

            hue += 1.0 / SourceBuffer.N_COLORS
            if hue > 1.0:
                hue = 0.0

        self.start_hue += 1.0 / SourceBuffer.N_COLORS
        if self.start_hue > 1.0:
            self.start_hue = 0.0

    def tag_event_handler(self, tag, widget, event, it):
        char_index = it.get_offset()
        tag_name = tag.get_property("name")
        if event.type == Gdk.MOTION_NOTIFY:
            print(f"Motion event at char {char_index} tag `{tag_name}'\n")
        elif event.type == Gdk.EventType.BUTTON_PRESS:
            print(f"Button press at char {char_index} tag `{tag_name}'\n")
        elif event.type == Gdk.EventType.TwoButtonPress:
            print(f"Double click at char {char_index} tag `{tag_name}'\n")
        elif event.type == Gdk.EventType.TRIPLE_BUTTON_PRESS:
            print(f"Triple click at char {char_index} tag `{tag_name}'\n")
        elif event.type == Gdk.BUTTON_RELEASE:
            print(f"Button release at char {char_index} tag `{tag_name}'\n")
        elif event.type == Gdk.KEY_PRESS or event.type == Gdk.KEY_RELEASE:
            print(f"Key event at char {char_index} tag `{tag_name}'\n")
        return False

    def open_file(self, filename):
        with open(filename, "r") as f:
            content = f.read()
            self.set_text(content)
            self.set_modified(False)
            self.filename = filename


class ShoebotFileChooserDialog(Gtk.FileChooserDialog):
    CWD = None

    def __init__(self, *args, **kwargs):
        super(ShoebotFileChooserDialog, self).__init__(*args, **kwargs)

        # set some defaults
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_property("do-overwrite-confirmation", True)

        # set the working directory if available
        if ShoebotFileChooserDialog.CWD is not None:
            self.set_current_folder(ShoebotFileChooserDialog.CWD)

    def run(self):
        response = super(ShoebotFileChooserDialog, self).run()

        # get the working directory if the user clicked accepted the action
        if response == Gtk.ResponseType.ACCEPT:
            ShoebotFileChooserDialog.CWD = self.get_current_folder()

        return response


class ConsoleWindow:
    def __init__(self):
        # we define a scrollable window with automatic behavior for scrolling bars
        self.text_window = Gtk.ScrolledWindow()
        self.text_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.text_area = Gtk.TextView()
        self.text_area.set_editable(False)
        self.text_area.set_wrap_mode(Gtk.WrapMode.WORD)
        self.text_area.connect("size-allocate", self.on_contents_changed)
        self.text_window.add(self.text_area)
        # here we set default values for background and text of console window
        self.text_area.modify_base(Gtk.StateType.NORMAL, Gdk.color_parse("dark grey"))
        self.text_area.modify_text(Gtk.StateType.NORMAL, Gdk.color_parse("red"))
        # then we define some text tag for defining colors for system messages and stdout
        self.tag_table = self.text_buffer.get_tag_table()

        self.stdout_tag = self.text_buffer.create_tag(
            "stdout", foreground="black", weight=600, size_points=9
        )
        self.system_message_tag = self.text_buffer.create_tag(
            "system", foreground="darkgrey"
        )
        self.text_area.modify_font(Pango.FontDescription("monospace 9"))

    def write(self, message, output=None, system=None):
        if not output:
            # no tags set for stderr messages, color will be the one set for TextView
            self.text_buffer.insert_at_cursor(message)
        elif system:
            # if output and system values are set, text is treated as a system message
            # and system tag is used
            it = self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert())
            self.text_buffer.insert_with_tags(it, message, self.system_message_tag)
        else:
            # if only output value is set, tag used will be stdout
            it = self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert())
            self.text_buffer.insert_with_tags(it, message, self.stdout_tag)

        while Gtk.events_pending():
            # Refresh the Window
            Gtk.main_iteration()

    def clear(self):
        self.text_buffer.set_text("")

    def on_contents_changed(self, widget, event):
        # scroll to bottom when there's new text
        # https://stackoverflow.com/questions/5218948/how-to-auto-scroll-a-gtk-scrolledwindow
        adj = self.text_window.get_vadjustment()
        adj.set_value(adj.get_upper() - adj.get_page_size())

    @property
    def text_buffer(self):
        return self.text_area.get_buffer()


class StdoutFilter(object):
    def __init__(self, parent):
        self.parent = parent

    def write(self, data):
        message = data
        self.parent.write(message, True)

    def flush(self):
        pass


UI_INFO = """
<ui>
  <menubar name='MenuBar'>
    <menu action='FileMenu'>
      <menuitem action='FileNew' />
      <menuitem action='FileOpen' />
      <menuitem action='FileSave' />
      <menuitem action='FileSaveAs' />
      <separator />
      <menuitem action='FileClose' />
      <menuitem action='FileQuit' />
    </menu>
    <menu action='EditMenu'>
      <menuitem action='EditUndo' />
      <menuitem action='EditRedo' />
      <separator />
      <menuitem action='EditFind' />
      <separator />
      <menuitem action='ClearConsole' />
    </menu>
    <menu action='RunMenu'>
      <menuitem action='Run' />
      <separator />
      <menuitem action='VarWindow' />
      <menuitem action='FullScreen' />
      <menuitem action='SocketServer' />
    </menu>
    <menu action='SettingsMenu'>
      <menuitem action='WrapNone' />
      <menuitem action='WrapWords' />
      <menuitem action='WrapChars' />
      <separator />
      <menuitem action='ThemeLight' />
      <menuitem action='ThemeDark' />
    </menu>
    <menu action='HelpMenu'>
      <menuitem action='HelpAbout' />
    </menu>
  </menubar>
</ui>
"""


class ShoebotEditorWindow(Gtk.Window):
    FONT = None

    def __init__(self, filename=None):
        GObject.GObject.__init__(self)
        ShoebotIDE.add_editor_window(self)

        source_buffer = SourceBuffer(filename)
        source_buffer.ref()

        self.connect("delete_event", self.on_close_window)

        action_group = Gtk.ActionGroup("menubar_actions")
        action_group.add_actions(
            [
                ("FileMenu", None, _("_File")),
                (
                    "FileNew",
                    Gtk.STOCK_NEW,
                    _("_New"),
                    "<control>N",
                    None,
                    self.on_new_file,
                ),
                (
                    "FileOpen",
                    Gtk.STOCK_OPEN,
                    _("_Open"),
                    "<control>O",
                    None,
                    self.on_open_file,
                ),
                (
                    "FileSave",
                    Gtk.STOCK_SAVE,
                    _("_Save"),
                    "<control>S",
                    None,
                    self.on_save_file,
                ),
                (
                    "FileSaveAs",
                    Gtk.STOCK_SAVE_AS,
                    _("Save _As"),
                    "<control><alt>S",
                    None,
                    self.on_save_file_as,
                ),
                (
                    "FileClose",
                    Gtk.STOCK_CLOSE,
                    _("_Close"),
                    "<control>W",
                    None,
                    self.on_close_file,
                ),
                (
                    "FileQuit",
                    Gtk.STOCK_QUIT,
                    _("_Quit"),
                    "<control>Q",
                    None,
                    self.on_quit,
                ),
            ]
        )

        action_group.add_actions(
            [
                ("EditMenu", None, _("_Edit")),
                (
                    "EditUndo",
                    Gtk.STOCK_UNDO,
                    _("_Undo"),
                    "<control>Z",
                    None,
                    self.on_undo,
                ),
                (
                    "EditRedo",
                    Gtk.STOCK_REDO,
                    _("_Redo"),
                    "<control><shift>Z",
                    None,
                    self.on_redo,
                ),
                (
                    "EditFind",
                    Gtk.STOCK_FIND,
                    _("_Find"),
                    "<control>F",
                    None,
                    self.on_find,
                ),
                (
                    "ClearConsole",
                    Gtk.STOCK_CLEAR,
                    _("_Clear console"),
                    "<control><shift>C",
                    None,
                    self.on_clear_console,
                ),
            ]
        )

        action_group.add_action(Gtk.Action("SettingsMenu", _("_Settings"), None, None))
        action_group.add_radio_actions(
            [
                ("WrapNone", None, _("Wrap None"), None, None, Gtk.WrapMode.NONE),
                ("WrapWords", None, _("Wrap Words"), None, None, Gtk.WrapMode.WORD),
                ("WrapChars", None, _("Wrap Chars"), None, None, Gtk.WrapMode.CHAR),
            ],
            1,
            self.on_wrap_changed,
        )
        action_group.add_radio_actions(
            [
                ("ThemeLight", None, _("Light Theme"), None, None, True),
                ("ThemeDark", None, _("Dark Theme"), None, None, False),
            ],
            0 if ShoebotIDE.dark_theme else 1,
            self.on_theme_changed,
        )

        action_group.add_actions(
            [
                ("RunMenu", None, _("_Run")),
                (
                    _("Run"),
                    Gtk.STOCK_MEDIA_PLAY,
                    "_Run Script",
                    "<control>R",
                    None,
                    self.on_run_script,
                ),
            ]
        )
        variable_window_action = Gtk.ToggleAction(
            "VarWindow", _("Show variables window"), None, None
        )
        variable_window_action.connect("toggled", self.on_varwindow_changed)
        action_group.add_action(variable_window_action)
        full_screen_action = Gtk.ToggleAction(
            "FullScreen", _("Full screen"), None, None
        )
        full_screen_action.connect("toggled", self.on_fullscreen_changed)
        action_group.add_action(full_screen_action)
        socket_server_action = Gtk.ToggleAction(
            "SocketServer", _("Run socket server"), None, None
        )
        socket_server_action.connect("toggled", self.on_socketserver_changed)
        action_group.add_action(socket_server_action)

        action_group.add_actions(
            [
                ("HelpMenu", None, _("_Help")),
                ("HelpAbout", Gtk.STOCK_INFO, _("_About"), None, None, self.on_about),
            ]
        )

        ui_manager = Gtk.UIManager()
        # Throws exception if something went wrong
        ui_manager.add_ui_from_string(UI_INFO)
        # Add the accelerator group to the toplevel window
        accel_group = ui_manager.get_accel_group()
        self.add_accel_group(accel_group)
        # Add menu actions
        ui_manager.insert_action_group(action_group)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        menu_bar = ui_manager.get_widget("/MenuBar")
        box.pack_start(menu_bar, False, False, 0)

        hpaned = Gtk.HPaned()
        vbox = Gtk.VBox(False, 0)
        hpaned.add1(vbox)
        box.pack_start(hpaned, True, True, 0)

        self.add(box)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        self.source_view = GtkSource.View.new_with_buffer(source_buffer)
        self.source_view.set_wrap_mode(Gtk.WrapMode.WORD)
        self.source_view.set_show_line_numbers(True)

        # Configure source view for editing python:
        self.source_view.set_auto_indent(True)
        self.source_view.set_insert_spaces_instead_of_tabs(True)
        self.source_view.set_tab_width(4)
        self.source_view.set_indent_width(4)
        # self.source_view.connect("expose_event", self.tab_stops_expose)

        self.toggle_dark_theme(dark=ShoebotIDE.dark_theme)

        self.bhid = source_buffer.connect("mark_set", self.cursor_set_callback)

        if ShoebotEditorWindow.FONT is None:
            # Get font or fallback
            context = self.source_view.get_pango_context()
            fonts = context.list_families()
            for font in fonts:
                if font.get_name() == "Bitstream Vera Sans Mono":
                    ShoebotEditorWindow.FONT = "Bitstream Vera Sans Mono 10"
                    break
            else:
                print("Bitstream Vera Font not found.")
                print("Download and install it from here")
                print("http://ftp.gnome.org/pub/GNOME/sources/ttf-bitstream-vera/1.10/")
                ShoebotEditorWindow.FONT = "Mono 10"

            self.source_view.modify_font(
                Pango.FontDescription(ShoebotEditorWindow.FONT)
            )

        vbox.pack_start(scrolled_window, True, True, 0)
        scrolled_window.add(self.source_view)

        # this creates a console error and output window besides script window
        self.console_error = ConsoleWindow()
        # we create an instance for stdout filter
        self.stdout_filter = StdoutFilter(self.console_error)
        # we redirect stderr
        if not DEBUG:
            sys.stderr = self.console_error
            # stdout is redirected too, but through the filter in order to get different color for text
            sys.stdout = self.stdout_filter
            # error-console window is added to container as second child
        hpaned.add2(self.console_error.text_window)
        hpaned.set_position(450)
        # message displayed in console-error window at start, the double true values passed makes it render with system message tag
        self.console_error.write(
            _(
                "This is the console window.\n\nScript output and error messages are shown here.\n\nYou can clear the window with the 'Edit - Clear console' option or pressing Ctrl-Shift-C.\n\n"
            ),
            True,
            True,
        )

        self.set_default_size(800, 500)
        self.source_view.grab_focus()

        self.update_window_title()
        self.init_menus()

        self.shoebot_window = None

        # option toggles
        self.use_varwindow = False
        self.use_socketserver = False
        self.go_fullscreen = False

        # setup syntax highlighting
        language_manager = GtkSource.LanguageManager()
        language = language_manager.guess_language(None, "text/x-python")
        source_buffer.set_language(language)

        try:
            self.set_icon_from_file(ICON_FILE)
        except GObject.GError:
            # icon not found = no icon
            pass

        self.show_all()

    @staticmethod
    def open_file(filename):
        """
        Open filename.

        :param filename: filename to open.
        :return:  True if file was opened.
        """
        try:
            ShoebotEditorWindow(filename)
            return True
        except IOError as e:
            errmsg = e.args[1]
            dialog = Gtk.MessageDialog(
                None,
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.INFO,
                Gtk.ButtonsType.OK,
                _("Cannot open file '%s': %s") % (filename, errmsg),
            )
        dialog.run()
        dialog.destroy()
        return False

    def on_close_file(self, widget):
        if self.confirm_quit_or_save():
            self.close_view()
            return False
        return True

    def on_close_window(self, window, event, data=None):
        if self.confirm_quit_or_save():
            self.close_view()
            return False
        return True

    def on_new_file(self, widget):
        ShoebotEditorWindow()

    def on_open_file(self, widget):
        chooser = ShoebotFileChooserDialog(
            _("Open File"),
            None,
            Gtk.FileChooserAction.OPEN,
            (
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.ACCEPT,
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
            ),
        )

        if chooser.run() == Gtk.ResponseType.ACCEPT:
            try:
                source_buffer = self.get_source_buffer()
                if (
                    source_buffer.can_undo() is False
                    and source_buffer.can_redo() is False
                ):
                    # Opening in an unmodified buffer, so a new window is not needed
                    source_buffer.open_file(chooser.get_filename())
                else:
                    # Default, open a new window
                    ShoebotEditorWindow(chooser.get_filename())
            except IOError as e:
                errmsg = e.args[1]
                error_dialog = Gtk.MessageDialog(
                    None,
                    Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.INFO,
                    Gtk.ButtonsType.OK,
                    f"Cannot open file '{chooser.get_filename()}': {errmsg}",
                )
                error_dialog.run()
                error_dialog.destroy()

        chooser.destroy()

    def save_or_save_as(self):
        """
        Call save_as if file has never been saved otherwise call save.
        """
        filename = self.source_view.get_buffer().filename
        if filename:
            self.save(filename)
        else:
            self.save_as()

    def on_save_file_as(self, widget):
        self.save_as()

    def on_save_file(self, widget):
        self.save_or_save_as()

    def on_quit(self, widget):
        for view in ShoebotIDE.editor_windows:
            if not view.confirm_quit_or_save():
                return
        if self.shoebot_window is not None:
            self.shoebot_window.destroy()

        Gtk.main_quit()
        sys.exit()

    def on_insert_and_scroll(self, callback_action, widget):
        buffer = self.get_source_buffer()

        start, end = buffer.get_bounds()
        mark = buffer.create_mark(None, end, False)

        buffer.insert(
            end,
            "Hello this is multiple lines of text\n"
            "Line 1\n"
            "Line 2\n"
            "Line 3\n"
            "Line 4\n"
            "Line 5\n",
        )

        self.source_view.scroll_to_mark(mark, 0, True, 0.0, 1.0)
        buffer.delete_mark(mark)

    def toggle_dark_theme(self, dark=False):
        Gtk.Settings.get_default().set_property(
            "gtk-application-prefer-dark-theme", dark
        )
        if dark:
            scheme_name = "cobalt"
        else:
            scheme_name = "classic"

        ShoebotIDE.set_source_buffers_style_scheme(scheme_name)
        ShoebotIDE.dark_theme = dark

    def on_theme_changed(self, widget, current):
        # Theme is either ThemeLight or ThemeDark
        self.toggle_dark_theme(dark=current.get_name() == "ThemeDark")

    def on_wrap_changed(self, widget, current):
        self.source_view.set_wrap_mode(GTK_WRAP_MODES[current.get_name()])

    def on_varwindow_changed(self, widget):
        self.use_varwindow = widget.get_active()

    def on_socketserver_changed(self, widget):
        self.use_socketserver = widget.get_active()

    def on_fullscreen_changed(self, widget):
        self.go_fullscreen = widget.get_active()

    def on_color_cycle_changed(self, widget):
        self.source_view.get_buffer().set_colors(widget.get_active())

    def on_apply_tabs(self, callback_action, widget):
        source_buffer = self.get_source_buffer()
        bounds = source_buffer.get_selection_bounds()
        if bounds:
            start, end = bounds
            if callback_action:
                source_buffer.remove_tag(source_buffer.custom_tabs_tag, start, end)
            else:
                source_buffer.apply_tag(source_buffer.custom_tabs_tag, start, end)

    def on_apply_colors(self, callback_action, widget):
        source_buffer = self.get_source_buffer()
        bounds = source_buffer.get_selection_bounds()
        if bounds:
            start, end = bounds
            if not callback_action:
                for tag in source_buffer.color_tags:
                    source_buffer.remove_tag(tag, start, end)
            else:
                tmp = source_buffer.color_tags
                i = 0
                next = start.copy()
                while next.compare(end) < 0:
                    next.forward_chars(2)
                    if next.compare(end) >= 0:
                        next = end

                    source_buffer.apply_tag(tmp[i], start, next)
                    i += 1
                    if i >= len(tmp):
                        i = 0
                    start = next.copy()

    def on_remove_tags(self, callback_action, widget):
        source_buffer = self.get_source_buffer()
        bounds = source_buffer.get_selection_bounds()
        if bounds:
            start, end = bounds
            source_buffer.remove_all_tags(start, end)

    def on_clear_console(self, widget):
        self.console_error.clear()

    def search_dialog_handler(self, dialog, search_direction):
        if search_direction not in [SEARCH_FORWARD, SEARCH_BACKWARD]:
            dialog.destroy()
            return

        start, end = dialog.get_source_buffer.get_bounds()
        search_string = start.get_text(end)

        print(_(f"Searching for `{search_string}'\n"))

        buffer = self.get_source_buffer()
        buffer.search(search_string, self, search_direction)
        dialog.destroy()

    def on_find(self, widget):
        search_text = Gtk.TextView()
        dialog = Gtk.Dialog(
            _("Search"),
            self,
            Gtk.DialogFlags.DESTROY_WITH_PARENT,
            (
                _("Forward"),
                SEARCH_FORWARD,
                _("Backward"),
                SEARCH_BACKWARD,
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.NONE,
            ),
        )
        dialog.vbox.pack_end(search_text, True, True, 0)
        dialog.source_buffer = search_text.get_buffer()
        dialog.connect("response", self.search_dialog_handler)

        search_text.show()
        search_text.grab_focus()
        dialog.show_all()

    def on_undo(self, widget):
        source_buffer = self.get_source_buffer()
        if source_buffer.can_undo():
            source_buffer.undo()

    def on_redo(self, widget):
        source_buffer = self.get_source_buffer()
        if source_buffer.can_redo():
            source_buffer.redo()

    def on_about(self, widget):
        about_dialog = Gtk.AboutDialog()
        website = "http://shoebot.net/"
        authors = [
            "Dave Crossland <dave AT lab6.com>",
            "est <electronixtar AT gmail.com>",
            "Francesco Fantoni <francesco AT hv-a.com>",
            "Paulo Silva <nitrofurano AT gmail.com>",
            "Pedro Angelo <pangelo AT virii-labs.org>",
            "Ricardo Lafuente <ricardo AT sollec.org>",
            "Stuart Axon <stuaxo2 AT yahoo.com>",
            "Tetsuya Saito <t2psyto AT gmail.com>",
        ]
        about_dialog.set_version("1.4")
        about_dialog.set_name("shoebot")
        about_dialog.set_license("GPLv3")
        about_dialog.set_authors(authors)
        about_dialog.set_website(website)

        about_dialog.connect("response", lambda dialog, data: dialog.destroy())
        about_dialog.show_all()

    def init_menus(self):
        text_view = self.source_view
        direction = text_view.get_direction()
        wrap_mode = text_view.get_wrap_mode()
        menu_item = None

        # if direction == Gtk.TextDirection.LTR:
        #     menu_item = self.item_factory.get_widget("/Settings/Left-to-Right")
        # elif direction == Gtk.TextDirection.RTL:
        #     menu_item = self.item_factory.get_widget("/Settings/Right-to-Left")

        if menu_item:
            menu_item.activate()

        # if wrap_mode == Gtk.WrapMode.NONE:
        #     menu_item = self.item_factory.get_widget("/Settings/Wrap Off")
        # elif wrap_mode == Gtk.WrapMode.WORD:
        #     menu_item = self.item_factory.get_widget("/Settings/Wrap Words")
        # elif wrap_mode == Gtk.WrapMode.CHAR:
        #     menu_item = self.item_factory.get_widget("/Settings/Wrap Chars")

        # if menu_item:
        #     menu_item.activate()

    def close_view(self):
        ShoebotIDE.remove_editor_window(self)
        source_buffer = self.get_source_buffer()
        # source_buffer.unref()
        source_buffer.disconnect(self.bhid)
        self.source_view.destroy()
        del self.source_view
        self.source_view = None
        self.destroy()
        del self
        if not ShoebotIDE.editor_windows:
            Gtk.main_quit()

    def confirm_quit_or_save(self):
        if self.get_source_buffer().get_modified():
            dialog = Gtk.MessageDialog(
                self,
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.QUESTION,
                Gtk.ButtonsType.YES_NO,
                _("Save changes to '%s'?") % self.get_source_buffer().pretty_name(),
            )
            dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
            result = dialog.run()
            dialog.destroy()
            if result == Gtk.ResponseType.YES:
                self.save_or_save_as()
            else:
                return result == Gtk.ResponseType.NO
        return True

    def save(self, filename):
        """
        :param filename:  Filename to save to
        :return:  True on success.

        Updates source_buffer.filename on success.
        """
        save_succeeded = False
        have_backup = False
        backup_filename = filename + "~"
        try:
            if os.path.isfile(backup_filename):
                os.remove(backup_filename)
            os.rename(filename, backup_filename)
        except (OSError, IOError) as e:
            errnum, errmsg = e.args
            if errnum != errno.ENOENT:
                err = f"Cannot back up '{filename}' to '{backup_filename}': {errmsg}"
                dialog = Gtk.MessageDialog(
                    self,
                    Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.INFO,
                    Gtk.ButtonsType.OK,
                    err,
                )
                dialog.run()
                dialog.destroy()
                return False

        have_backup = True
        source_buffer = self.get_source_buffer()
        start, end = source_buffer.get_bounds()
        chars = source_buffer.get_slice(start, end, False)
        try:
            with open(self.get_source_buffer().filename, "w") as f:
                f.write(chars)
            save_succeeded = True
            self.get_source_buffer().set_modified(False)
            self.get_source_buffer().filename = filename
        except IOError as e:
            err = f"Error writing to '{filename}': {e.args[1]}"
            dialog = Gtk.MessageDialog(
                self,
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.INFO,
                Gtk.ButtonsType.OK,
                err,
            )
            dialog.run()
            dialog.destroy()

        if not save_succeeded and have_backup:
            try:
                os.rename(backup_filename, filename)
            except OSError as e:
                err = f"Can't restore backup file '{filename}' to '{back_filename}': {e.args[1]}\nBackup left as '{backup_filename}'"
                dialog = Gtk.MessageDialog(
                    self,
                    Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.INFO,
                    Gtk.ButtonsType.OK,
                    err,
                )
                dialog.run()
                dialog.destroy()

        return save_succeeded

    def save_as(self):
        """
        Return True if the source_buffer was saved
        """
        chooser = ShoebotFileChooserDialog(
            _("Save File"),
            None,
            Gtk.FileChooserAction.SAVE,
            (
                Gtk.STOCK_SAVE,
                Gtk.ResponseType.ACCEPT,
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
            ),
        )
        chooser.set_do_overwrite_confirmation(True)
        chooser.set_transient_for(self)
        saved = chooser.run() == Gtk.ResponseType.ACCEPT
        if saved:
            filename = chooser.get_filename()
            if self.save(filename):
                self.update_window_title()
        chooser.destroy()
        return saved

    def update_window_title(self):
        """
        Update the window title from the filename in the source_buffer.
        :return:
        """
        save_prefix = "*" if self.source_view.get_buffer().get_modified() else ""
        window_title = (
            f"Shoebot - {save_prefix}{self.source_view.get_buffer().pretty_name()}"
        )
        self.set_title(window_title)

    def cursor_set_callback(self, buffer, location, mark):

        # TODO: here should go the line syntax highlighter
        # 1. get buffer
        #      get modification state
        # 2. get line contents
        # 3. replace by pygmentised content
        #      revert to modification state
        self.update_window_title()

    def tab_stops_expose(self, widget, event):
        # print(self, widget, event)
        text_view = widget

        # See if this expose is on the tab stop window
        top_win = text_view.get_window(Gtk.TextWindowType.TOP)
        bottom_win = text_view.get_window(Gtk.TextWindowType.BOTTOM)

        if event.window == top_win:
            type = Gtk.TextWindowType.TOP
            target = top_win
        elif event.window == bottom_win:
            type = Gtk.TextWindowType.BOTTOM
            target = bottom_win
        else:
            return False

        first_x = event.area.x
        last_x = first_x + event.area.width

        first_x, y = text_view.window_to_buffer_coords(type, first_x, 0)
        last_x, y = text_view.window_to_buffer_coords(type, last_x, 0)

        source_buffer = text_view.get_buffer()
        insert = source_buffer.get_iter_at_mark(source_buffer.get_insert())
        attrs = Gtk.TextAttributes()
        insert.get_attributes(attrs)

        tabslist = []
        in_pixels = False
        if attrs.tabs:
            tabslist = attrs.tabs.get_tabs()
            in_pixels = attrs.tabs.get_positions_in_pixels()

        for align, position in tabslist:
            if not in_pixels:
                position = Pango.PIXELS(position)

            pos, y = text_view.buffer_to_window_coords(type, position, 0)
            target.draw_line(text_view.style.fg_gc[text_view.state], pos, 0, pos, 15)

        return True

    def get_lines(self, first_y, last_y, buffer_coords, numbers):
        text_view = self.source_view
        # Get iter at first y
        it, top = text_view.get_line_at_y(first_y)

        # For each iter, get its location and add it to the arrays.
        # Stop when we pass last_y
        count = 0

        while not it.is_end():
            y, height = text_view.get_line_yrange(it)
            buffer_coords.append(y)
            line_num = it.get_line()
            numbers.append(line_num)
            count += 1
            if (y + height) >= last_y:
                break
            it.forward_line()

        return count

    def on_run_script(self, widget):
        # get the buffer contents
        source_buffer = self.source_view.get_buffer()
        start, end = source_buffer.get_bounds()
        codestring = source_buffer.get_text(start, end, include_hidden_chars=False)
        window_title = self.get_title()
        try:
            buffer_dir = os.path.dirname(source_buffer.filename or "")
            if buffer_dir:
                os.chdir(buffer_dir)

            bot = shoebot.create_bot(
                codestring,
                shoebot.NODEBOX,
                server=self.use_socketserver,
                show_vars=self.use_varwindow,
                title=window_title,
                window=True,
            )
            self.shoebot_window = bot._canvas.sink
            bot.run(
                codestring, run_forever=True, max_iterations=None, frame_limiter=True
            )
        except (ShoebotError, NameError):
            import traceback
            import sys

            errmsg = traceback.format_exc(limit=1)
            err = _("Error in Shoebot script:") + "\n {errmsg}"
            dialog = Gtk.MessageDialog(
                self,
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.INFO,
                Gtk.ButtonsType.OK,
                err,
            )
            result = dialog.run()
            dialog.destroy()
            self.shoebot_window = None
            return False

        # TODO: have a try/except that shows an error window

    def get_source_buffer(self):
        return self.source_view.get_buffer()


class ShoebotIDE:
    colormap = None
    untitled_file_counter = 0
    editor_windows = list()

    dark_theme = (
        Gtk.Settings.get_default().get_property("gtk-theme-name").endswith("-dark")
    )  # TODO - Is there a proper way of doing this?

    def __init__(self, filenames):
        if not filenames:
            ShoebotEditorWindow()
        else:
            self.open_files(filenames)

    @classmethod
    def open_files(cls, filenames):
        files_not_opened = []
        files_were_opened = False
        for filename in filenames:
            filename = os.path.abspath(filename)
            try:
                ShoebotEditorWindow(filename)
            except IOError as e:
                files_not_opened.append(filename)
            else:
                files_were_opened = True

        if files_not_opened:
            dialog = Gtk.MessageDialog(
                None,
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.INFO,
                Gtk.ButtonsType.OK,
                _("Cannot open files:") + "\n%s" % "\n".join(files_not_opened),
            )
            dialog.run()
            dialog.destroy()

        if not files_were_opened:
            sys.exit(1)

    @classmethod
    def set_source_buffers_style_scheme(cls, scheme_name):
        style_manager = GtkSource.StyleSchemeManager.get_default()
        scheme = style_manager.get_scheme(scheme_name)

        for view in cls.editor_windows:
            source_buffer = view.get_source_buffer()
            source_buffer.set_style_scheme(scheme)

    @classmethod
    def add_editor_window(cls, view):
        cls.editor_windows.append(view)

    @classmethod
    def remove_editor_window(cls, view):
        cls.editor_windows.remove(view)

    @classmethod
    def get_next_untitled_filename(cls):
        cls.untitled_file_counter += 1
        if cls.untitled_file_counter == 1:
            return _("Untitled")
        else:
            return _("Untitled #%d") % cls.untitled_file_counter

    @property
    def source_buffer(self):
        return self.source_view.get_buffer()


def main():
    parser = argparse.ArgumentParser(_("usage: shoebot [file...[file]]"))
    parser.add_argument("filenames", nargs="*")

    args = parser.parse_args()

    filenames = vars(args)["filenames"]

    app = ShoebotIDE(filenames)
    Gtk.main()


if __name__ == "__main__":
    main()
